import itertools
import logging
import math
from collections.abc import Iterator
from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import toolz
import zarr
from toolz import map

from cubed.backend_array_api import (
    backend_array_to_numpy_array,
    numpy_array_to_backend_array,
)
from cubed.runtime.types import CubedPipeline
from cubed.storage.zarr import T_ZarrArray, lazy_zarr_array
from cubed.types import T_Chunks, T_DType, T_Shape, T_Store
from cubed.utils import (
    array_memory,
    chunk_memory,
    get_item,
    map_nested,
    split_into,
    to_chunksize,
)
from cubed.vendor.dask.array.core import normalize_chunks
from cubed.vendor.dask.blockwise import _get_coord_mapping, _make_dims, lol_product
from cubed.vendor.dask.core import flatten

from .types import CubedArrayProxy, MemoryModeller, PrimitiveOperation

logger = logging.getLogger(__name__)


sym_counter = 0


def gensym(name: str) -> str:
    global sym_counter
    sym_counter += 1
    return f"{name}-{sym_counter:03}"


@dataclass(frozen=True)
class BlockwiseSpec:
    """Specification for how to run blockwise on an array.

    This is similar to ``CopySpec`` in rechunker.

    Attributes
    ----------
    key_function : Callable
        A function that maps an output chunk key to one or more input chunk keys.
    function : Callable
        A function that maps input chunks to an output chunk.
    function_nargs: int
        The number of array arguments that ``function`` takes.
    num_input_blocks: Tuple[int, ...]
        The number of input blocks read from each input array.
    reads_map : Dict[str, CubedArrayProxy]
        Read proxy dictionary keyed by array name.
    write : CubedArrayProxy
        Write proxy with an ``array`` attribute that supports ``__setitem__``.
    """

    key_function: Callable[..., Any]
    function: Callable[..., Any]
    function_nargs: int
    num_input_blocks: Tuple[int, ...]
    reads_map: Dict[str, CubedArrayProxy]
    write: CubedArrayProxy


def apply_blockwise(out_coords: List[int], *, config: BlockwiseSpec) -> None:
    """Stage function for blockwise."""
    # lithops needs params to be lists not tuples, so convert back
    out_coords_tuple = tuple(out_coords)
    out_chunk_key = key_to_slices(
        out_coords_tuple, config.write.array, config.write.chunks
    )

    # get array chunks for input keys, preserving any nested list structure
    args = []
    get_chunk_config = partial(get_chunk, config=config)
    out_key = ("out",) + out_coords_tuple  # array name is ignored by key_function
    in_keys = config.key_function(out_key)
    for in_key in in_keys:
        arg = map_nested(get_chunk_config, in_key)
        args.append(arg)

    result = config.function(*args)
    if isinstance(result, dict):  # structured array with named fields
        for k, v in result.items():
            v = backend_array_to_numpy_array(v)
            config.write.open().set_basic_selection(out_chunk_key, v, fields=k)
    else:
        result = backend_array_to_numpy_array(result)
        config.write.open()[out_chunk_key] = result


def key_to_slices(
    key: Tuple[int, ...], arr: T_ZarrArray, chunks: Optional[T_Chunks] = None
) -> Tuple[slice, ...]:
    """Convert a chunk index key to a tuple of slices"""
    chunks = normalize_chunks(chunks or arr.chunks, shape=arr.shape, dtype=arr.dtype)
    return get_item(chunks, key)


def get_chunk(in_key, config):
    """Read a chunk from the named array"""
    name = in_key[0]
    in_coords = in_key[1:]
    arr = config.reads_map[name].open()
    selection = key_to_slices(in_coords, arr)
    arg = arr[selection]
    arg = numpy_array_to_backend_array(arg)
    return arg


def blockwise(
    func: Callable[..., Any],
    out_ind: Sequence[Union[str, int]],
    *args: Any,
    allowed_mem: int,
    reserved_mem: int,
    target_store: T_Store,
    target_path: Optional[str] = None,
    storage_options: Optional[Dict[str, Any]] = None,
    shape: T_Shape,
    dtype: T_DType,
    chunks: T_Chunks,
    new_axes: Optional[Dict[int, int]] = None,
    in_names: Optional[List[str]] = None,
    out_name: Optional[str] = None,
    extra_projected_mem: int = 0,
    extra_func_kwargs: Optional[Dict[str, Any]] = None,
    fusable: bool = True,
    num_input_blocks: Optional[Tuple[int, ...]] = None,
    **kwargs,
):
    """Apply a function to multiple blocks from multiple inputs, expressed using concise indexing rules.

    Unlike ```general_blockwise``, an index notation is used to specify the block mapping,
    like in Dask Array.

    Parameters
    ----------
    func : callable
        Function to apply to individual tuples of blocks
    out_ind : iterable
        Block pattern of the output, something like 'ijk' or (1, 2, 3)
    *args : sequence of Array, index pairs
        Sequence like (x, 'ij', y, 'jk', z, 'i')
    allowed_mem : int
        The memory available to a worker for running a task, in bytes. Includes ``reserved_mem``.
    reserved_mem : int
        The memory reserved on a worker for non-data use when running a task, in bytes
    target_store : string or zarr.Array
        Path to output Zarr store, or Zarr array
    shape : tuple
        The shape of the output array.
    dtype : np.dtype
        The ``dtype`` of the output array.
    chunks : tuple
        The chunks of the output array.
    new_axes : dict
        New indexes and their dimension lengths
    extra_projected_mem : int
        Extra memory projected to be needed (in bytes) in addition to the memory used reading
        the input arrays and writing the output.
    extra_func_kwargs : dict
        Extra keyword arguments to pass to function that can't be passed as regular keyword arguments
        since they clash with other blockwise arguments (such as dtype).
    **kwargs : dict
        Extra keyword arguments to pass to function

    Returns
    -------
    CubedPipeline to run the operation
    """

    arrays: Sequence[T_ZarrArray] = args[::2]
    array_names = in_names or [f"in_{i}" for i in range(len(arrays))]

    inds: Sequence[Union[str, int]] = args[1::2]

    numblocks: Dict[str, Tuple[int, ...]] = {}
    for name, array in zip(array_names, arrays):
        input_chunks = normalize_chunks(
            array.chunks, shape=array.shape, dtype=array.dtype
        )
        numblocks[name] = tuple(map(len, input_chunks))

    argindsstr: List[Any] = []
    for name, ind in zip(array_names, inds):
        argindsstr.extend((name, ind))

    key_function = make_blockwise_key_function_flattened(
        func,
        out_name or "out",
        out_ind,
        *argindsstr,
        numblocks=numblocks,
        new_axes=new_axes,
    )

    return general_blockwise(
        func,
        key_function,
        *arrays,
        allowed_mem=allowed_mem,
        reserved_mem=reserved_mem,
        target_store=target_store,
        target_path=target_path,
        storage_options=storage_options,
        shape=shape,
        dtype=dtype,
        chunks=chunks,
        in_names=in_names,
        extra_projected_mem=extra_projected_mem,
        extra_func_kwargs=extra_func_kwargs,
        fusable=fusable,
        num_input_blocks=num_input_blocks,
        **kwargs,
    )


def general_blockwise(
    func: Callable[..., Any],
    key_function: Callable[..., Any],
    *arrays: Any,
    allowed_mem: int,
    reserved_mem: int,
    target_store: T_Store,
    target_path: Optional[str] = None,
    storage_options: Optional[Dict[str, Any]] = None,
    shape: T_Shape,
    dtype: T_DType,
    chunks: T_Chunks,
    in_names: Optional[List[str]] = None,
    extra_projected_mem: int = 0,
    extra_func_kwargs: Optional[Dict[str, Any]] = None,
    fusable: bool = True,
    num_input_blocks: Optional[Tuple[int, ...]] = None,
    **kwargs,
):
    """A more general form of ``blockwise`` that uses a function to specify the block
    mapping, rather than an index notation.

    Parameters
    ----------
    func : callable
        Function to apply to individual tuples of blocks
    key_function : callable
        A function that maps an output chunk key to one or more input chunk keys.
    *arrays : sequence of Array
        The input arrays.
    allowed_mem : int
        The memory available to a worker for running a task, in bytes. Includes ``reserved_mem``.
    reserved_mem : int
        The memory reserved on a worker for non-data use when running a task, in bytes
    target_store : string or zarr.Array
        Path to output Zarr store, or Zarr array
    shape : tuple
        The shape of the output array.
    dtype : np.dtype
        The ``dtype`` of the output array.
    chunks : tuple
        The chunks of the output array.
    extra_projected_mem : int
        Extra memory projected to be needed (in bytes) in addition to the memory used reading
        the input arrays and writing the output.
    extra_func_kwargs : dict
        Extra keyword arguments to pass to function that can't be passed as regular keyword arguments
        since they clash with other blockwise arguments (such as dtype).
    **kwargs : dict
        Extra keyword arguments to pass to function

    Returns
    -------
    PrimitiveOperation to run the operation
    """
    array_names = in_names or [f"in_{i}" for i in range(len(arrays))]
    array_map = {name: array for name, array in zip(array_names, arrays)}

    chunks = normalize_chunks(chunks, shape=shape, dtype=dtype)
    chunksize = to_chunksize(chunks)
    if isinstance(target_store, zarr.Array):
        target_array = target_store
    else:
        target_array = lazy_zarr_array(
            target_store,
            shape,
            dtype,
            chunks=chunksize,
            path=target_path,
            storage_options=storage_options,
        )

    func_kwargs = extra_func_kwargs or {}
    func_with_kwargs = partial(func, **{**kwargs, **func_kwargs})
    num_input_blocks = num_input_blocks or (1,) * len(arrays)
    read_proxies = {
        name: CubedArrayProxy(array, array.chunks) for name, array in array_map.items()
    }
    write_proxy = CubedArrayProxy(target_array, chunksize)
    spec = BlockwiseSpec(
        key_function,
        func_with_kwargs,
        len(arrays),
        num_input_blocks,
        read_proxies,
        write_proxy,
    )

    # calculate projected memory
    projected_mem = reserved_mem + extra_projected_mem
    # inputs
    for array in arrays:  # inputs
        # memory for a compressed and an uncompressed input array chunk
        # - we assume compression has no effect (so it's an overestimate)
        # - ideally we'd be able to look at nbytes_stored,
        #   but this is not possible in general since the array has not been written yet
        projected_mem += array_memory(array.dtype, array.chunks) * 2
    # output
    # memory for a compressed and an uncompressed output array chunk
    # - this assumes the blockwise function creates a new array)
    # - numcodecs uses a working output buffer that's the size of the array being compressed
    projected_mem += array_memory(dtype, chunksize) * 2

    if projected_mem > allowed_mem:
        raise ValueError(
            f"Projected blockwise memory ({projected_mem}) exceeds allowed_mem ({allowed_mem}), including reserved_mem ({reserved_mem})"
        )

    # this must be an iterator of lists, not of tuples, otherwise lithops breaks
    output_blocks = map(list, itertools.product(*[range(len(c)) for c in chunks]))
    num_tasks = math.prod(len(c) for c in chunks)

    pipeline = CubedPipeline(
        apply_blockwise,
        gensym("apply_blockwise"),
        output_blocks,
        spec,
    )
    return PrimitiveOperation(
        pipeline=pipeline,
        source_array_names=array_names,
        target_array=target_array,
        projected_mem=projected_mem,
        allowed_mem=allowed_mem,
        reserved_mem=reserved_mem,
        num_tasks=num_tasks,
        fusable=fusable,
    )


# Code for fusing blockwise operations


def is_fuse_candidate(primitive_op: PrimitiveOperation) -> bool:
    """
    Return True if a primitive operation is a candidate for blockwise fusion.
    """
    return primitive_op.pipeline.function == apply_blockwise


def can_fuse_primitive_ops(
    primitive_op1: PrimitiveOperation, primitive_op2: PrimitiveOperation
) -> bool:
    if is_fuse_candidate(primitive_op1) and is_fuse_candidate(primitive_op2):
        return primitive_op1.num_tasks == primitive_op2.num_tasks
    return False


def can_fuse_multiple_primitive_ops(
    name: str,
    primitive_op: PrimitiveOperation,
    predecessor_primitive_ops: List[PrimitiveOperation],
    *,
    max_total_num_input_blocks: Optional[int] = None,
) -> bool:
    if is_fuse_candidate(primitive_op) and all(
        is_fuse_candidate(p) for p in predecessor_primitive_ops
    ):
        # If the peak projected memory for running all the predecessor ops in
        # order is larger than allowed_mem then we can't fuse.
        peak_projected = peak_projected_mem(predecessor_primitive_ops)
        if peak_projected > primitive_op.allowed_mem:
            logger.debug(
                "can't fuse %s since peak projected memory for predecessor ops (%s) is greater than allowed (%s)",
                name,
                peak_projected,
                primitive_op.allowed_mem,
            )
            return False
        # If the number of input blocks for each input is not uniform, then we
        # can't fuse. (This should never happen since all operations are
        # currently uniform, and fused operations are too if fuse is applied in
        # topological order.)
        num_input_blocks = primitive_op.pipeline.config.num_input_blocks
        if not all(num_input_blocks[0] == n for n in num_input_blocks):
            logger.debug(
                "can't fuse %s since number of input blocks for each input is not uniform: %s",
                name,
                num_input_blocks,
            )
            return False
        if max_total_num_input_blocks is None:
            # If max total input blocks not specified, then only fuse if num
            # tasks of predecessor ops match.
            ret = all(
                primitive_op.num_tasks == p.num_tasks for p in predecessor_primitive_ops
            )
            if ret:
                logger.debug(
                    "can fuse %s since num tasks of predecessor ops match", name
                )
            else:
                logger.debug(
                    "can't fuse %s since num tasks of predecessor ops do not match",
                    name,
                )
            return ret
        else:
            total_num_input_blocks = 0
            for ni, p in zip(num_input_blocks, predecessor_primitive_ops):
                for nj in p.pipeline.config.num_input_blocks:
                    total_num_input_blocks += ni * nj
            ret = total_num_input_blocks <= max_total_num_input_blocks
            if ret:
                logger.debug(
                    "can fuse %s since total number of input blocks (%s) does not exceed max (%s)",
                    name,
                    total_num_input_blocks,
                    max_total_num_input_blocks,
                )
            else:
                logger.debug(
                    "can't fuse %s since total number of input blocks (%s) exceeds max (%s)",
                    name,
                    total_num_input_blocks,
                    max_total_num_input_blocks,
                )
            return ret
    logger.debug(
        "can't fuse %s since primitive op and predecessors are not all candidates", name
    )
    return False


def peak_projected_mem(primitive_ops):
    """Calculate the peak projected memory for running a series of primitive ops
    and retaining their return values in memory."""
    memory_modeller = MemoryModeller()
    for p in primitive_ops:
        memory_modeller.allocate(p.projected_mem)
        chunkmem = chunk_memory(p.target_array)
        memory_modeller.free(p.projected_mem - chunkmem)
    return memory_modeller.peak_mem


def fuse(
    primitive_op1: PrimitiveOperation, primitive_op2: PrimitiveOperation
) -> PrimitiveOperation:
    """
    Fuse two blockwise operations into a single operation, avoiding writing to (or reading from) the target of the first operation.
    """

    assert primitive_op1.num_tasks == primitive_op2.num_tasks

    pipeline1 = primitive_op1.pipeline
    pipeline2 = primitive_op2.pipeline

    mappable = pipeline2.mappable

    def fused_key_func(out_key):
        return pipeline1.config.key_function(*pipeline2.config.key_function(out_key))

    def fused_func(*args):
        return pipeline2.config.function(pipeline1.config.function(*args))

    function_nargs = pipeline1.config.function_nargs
    read_proxies = pipeline1.config.reads_map
    write_proxy = pipeline2.config.write
    num_input_blocks = tuple(
        n * pipeline2.config.num_input_blocks[0]
        for n in pipeline1.config.num_input_blocks
    )
    spec = BlockwiseSpec(
        fused_key_func,
        fused_func,
        function_nargs,
        num_input_blocks,
        read_proxies,
        write_proxy,
    )

    source_array_names = primitive_op1.source_array_names
    target_array = primitive_op2.target_array
    projected_mem = max(primitive_op1.projected_mem, primitive_op2.projected_mem)
    allowed_mem = primitive_op2.allowed_mem
    reserved_mem = primitive_op2.reserved_mem
    num_tasks = primitive_op2.num_tasks

    pipeline = CubedPipeline(
        apply_blockwise,
        gensym("fused_apply_blockwise"),
        mappable,
        spec,
    )
    return PrimitiveOperation(
        pipeline=pipeline,
        source_array_names=source_array_names,
        target_array=target_array,
        projected_mem=projected_mem,
        allowed_mem=allowed_mem,
        reserved_mem=reserved_mem,
        num_tasks=num_tasks,
        fusable=True,
    )


def fuse_multiple(
    primitive_op: PrimitiveOperation, *predecessor_primitive_ops: PrimitiveOperation
) -> PrimitiveOperation:
    """
    Fuse a blockwise operation and its predecessors into a single operation, avoiding writing to (or reading from) the targets of the predecessor operations.
    """

    pipeline = primitive_op.pipeline
    predecessor_pipelines = [
        primitive_op.pipeline if primitive_op is not None else None
        for primitive_op in predecessor_primitive_ops
    ]

    # if a predecessor has no primitive op then use 1 for nargs
    predecessor_funcs_nargs = [
        pipeline.config.function_nargs if pipeline is not None else 1
        for pipeline in predecessor_pipelines
    ]

    mappable = pipeline.mappable

    def apply_pipeline_key_func(pipeline, n_input_blocks, arg):
        if pipeline is None:
            return (arg,)
        if n_input_blocks == 1:
            assert isinstance(arg, tuple)
            return pipeline.config.key_function(arg)
        else:
            # more than one input block is being read from arg
            assert isinstance(arg, (list, Iterator))
            if isinstance(arg, list):
                return tuple(
                    list(item)
                    for item in zip(*(pipeline.config.key_function(a) for a in arg))
                )
            else:
                # Return iterators to avoid materializing all array blocks at
                # once.
                return tuple(
                    iter(list(item))
                    for item in zip(*(pipeline.config.key_function(a) for a in arg))
                )

    def fused_key_func(out_key):
        # this will change when multiple outputs are supported
        args = pipeline.config.key_function(out_key)
        # split all args to the fused function into groups, one for each predecessor function
        func_args = tuple(
            item
            for i, (p, a) in enumerate(zip(predecessor_pipelines, args))
            for item in apply_pipeline_key_func(
                p, pipeline.config.num_input_blocks[i], a
            )
        )
        return split_into(func_args, predecessor_funcs_nargs)

    def apply_pipeline_func(pipeline, n_input_blocks, *args):
        if pipeline is None:
            return args[0]
        if n_input_blocks == 1:
            ret = pipeline.config.function(*args)
        else:
            # More than one input block is being read from this group of args to primitive op.
            # Note that it is important that a list is not returned to avoid materializing all
            # array blocks at once.
            ret = map(lambda item: pipeline.config.function(*item), zip(*args))
        return ret

    def fused_func(*args):
        # args are grouped appropriately so they can be called by each predecessor function
        func_args = [
            apply_pipeline_func(p, pipeline.config.num_input_blocks[i], *a)
            for i, (p, a) in enumerate(zip(predecessor_pipelines, args))
        ]
        return pipeline.config.function(*func_args)

    fused_function_nargs = pipeline.config.function_nargs
    # ok to get num_input_blocks[0] since it is uniform (see check in can_fuse_multiple_primitive_ops)
    fused_num_input_blocks = tuple(
        pipeline.config.num_input_blocks[0] * n
        for n in itertools.chain(
            *(
                p.pipeline.config.num_input_blocks if p is not None else (1,)
                for p in predecessor_primitive_ops
            )
        )
    )
    read_proxies = dict(pipeline.config.reads_map)
    for p in predecessor_pipelines:
        if p is not None:
            read_proxies.update(p.config.reads_map)
    write_proxy = pipeline.config.write
    spec = BlockwiseSpec(
        fused_key_func,
        fused_func,
        fused_function_nargs,
        fused_num_input_blocks,
        read_proxies,
        write_proxy,
    )

    source_array_names = []
    for i, p in enumerate(predecessor_primitive_ops):
        if p is None:
            source_array_names.append(primitive_op.source_array_names[i])
        else:
            source_array_names.extend(p.source_array_names)
    target_array = primitive_op.target_array
    projected_mem = max(
        primitive_op.projected_mem,
        peak_projected_mem(p for p in predecessor_primitive_ops if p is not None),
    )
    allowed_mem = primitive_op.allowed_mem
    reserved_mem = primitive_op.reserved_mem
    num_tasks = primitive_op.num_tasks

    fused_pipeline = CubedPipeline(
        apply_blockwise,
        gensym("fused_apply_blockwise"),
        mappable,
        spec,
    )
    return PrimitiveOperation(
        pipeline=fused_pipeline,
        source_array_names=source_array_names,
        target_array=target_array,
        projected_mem=projected_mem,
        allowed_mem=allowed_mem,
        reserved_mem=reserved_mem,
        num_tasks=num_tasks,
        fusable=True,
    )


# blockwise key functions


def make_blockwise_key_function(
    func: Callable[..., Any],
    output: str,
    out_indices: Sequence[Union[str, int]],
    *arrind_pairs: Any,
    numblocks: Optional[Dict[str, Tuple[int, ...]]] = None,
    new_axes: Optional[Dict[int, int]] = None,
) -> Callable[[List[int]], Any]:
    """Make a function that is the equivalent of make_blockwise_graph."""

    if numblocks is None:
        raise ValueError("Missing required numblocks argument.")
    new_axes = new_axes or {}
    argpairs = list(toolz.partition(2, arrind_pairs))

    # Dictionary mapping {i: 3, j: 4, ...} for i, j, ... the dimensions
    dims = _make_dims(argpairs, numblocks, new_axes)

    # Generate the abstract "plan" before constructing
    # the actual graph
    (coord_maps, concat_axes, dummies) = _get_coord_mapping(
        dims,
        output,
        out_indices,
        numblocks,
        argpairs,
        False,
    )

    def key_function(out_key):
        out_coords = out_key[1:]

        # from Dask make_blockwise_graph
        deps = set()
        coords = out_coords + dummies
        args = []
        for cmap, axes, (arg, ind) in zip(coord_maps, concat_axes, argpairs):
            if ind is None:
                args.append(arg)
            else:
                arg_coords = tuple(coords[c] for c in cmap)
                if axes:
                    tups = lol_product((arg,), arg_coords)
                    deps.update(flatten(tups))
                else:
                    tups = (arg,) + arg_coords
                    deps.add(tups)
                args.append(tups)

        args.insert(0, func)
        val = tuple(args)
        # end from make_blockwise_graph

        return val

    return key_function


def make_blockwise_key_function_flattened(
    func: Callable[..., Any],
    output: str,
    out_indices: Sequence[Union[str, int]],
    *arrind_pairs: Any,
    numblocks: Optional[Dict[str, Tuple[int, ...]]] = None,
    new_axes: Optional[Dict[int, int]] = None,
) -> Callable[[List[int]], Any]:
    # TODO: make this a part of make_blockwise_key_function?
    key_function = make_blockwise_key_function(
        func, output, out_indices, *arrind_pairs, numblocks=numblocks, new_axes=new_axes
    )

    def blockwise_fn_flattened(out_key):
        in_keys = key_function(out_key)[1:]  # drop function in position 0
        # flatten (nested) lists indicating contraction
        if isinstance(in_keys[0], list):
            in_keys = list(flatten(in_keys))
        return in_keys

    return blockwise_fn_flattened



from typing import Optional, Sequence

from networkx import MultiDiGraph

from cubed.runtime.pipeline import visit_nodes
from cubed.runtime.types import Callback, CubedPipeline, DagExecutor, TaskEndEvent
from cubed.runtime.utils import handle_operation_start_callbacks
from cubed.spec import Spec


def exec_stage_func(input, func=None, config=None, name=None, compute_id=None):
    return func(input, config=config)


class PythonDagExecutor(DagExecutor):
    """The default execution engine that runs tasks sequentially uses Python loops."""

    @property
    def name(self) -> str:
        return "single-threaded"

    def execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        resume: Optional[bool] = None,
        spec: Optional[Spec] = None,
        compute_id: Optional[str] = None,
        **kwargs,
    ) -> None:
        for name, node in visit_nodes(dag, resume=resume):
            handle_operation_start_callbacks(callbacks, name)
            pipeline: CubedPipeline = node["pipeline"]
            for m in pipeline.mappable:
                exec_stage_func(
                    m,
                    pipeline.function,
                    config=pipeline.config,
                    name=name,
                    compute_id=compute_id,
                )
                if callbacks is not None:
                    event = TaskEndEvent(name=name)
                    [callback.on_task_end(event) for callback in callbacks]



import asyncio
import multiprocessing
import os
from concurrent.futures import Executor, ProcessPoolExecutor, ThreadPoolExecutor
from functools import partial
from typing import Any, AsyncIterator, Callable, Iterable, Optional, Sequence

import cloudpickle
from aiostream import stream
from aiostream.core import Stream
from networkx import MultiDiGraph
from tenacity import Retrying, stop_after_attempt

from cubed.runtime.executors.asyncio import async_map_unordered
from cubed.runtime.pipeline import visit_node_generations, visit_nodes
from cubed.runtime.types import Callback, CubedPipeline, DagExecutor
from cubed.runtime.utils import (
    execution_stats,
    handle_callbacks,
    handle_operation_start_callbacks,
)
from cubed.spec import Spec


@execution_stats
def run_func(input, func=None, config=None, name=None, compute_id=None):
    print(f"{compute_id} {name}: running on {input}")
    result = func(input, config=config)
    return result


def unpickle_and_call(f, inp, **kwargs):
    import cloudpickle

    f = cloudpickle.loads(f)
    inp = cloudpickle.loads(inp)
    kwargs = {k: cloudpickle.loads(v) for k, v in kwargs.items()}
    return f(inp, **kwargs)


async def map_unordered(
    concurrent_executor: Executor,
    function: Callable[..., Any],
    input: Iterable[Any],
    retries: int = 2,
    use_backups: bool = True,
    batch_size: Optional[int] = None,
    return_stats: bool = False,
    name: Optional[str] = None,
    **kwargs,
) -> AsyncIterator[Any]:
    if retries == 0:
        retrying_function = function
    else:
        if isinstance(concurrent_executor, ProcessPoolExecutor):
            raise NotImplementedError("Retries not supported for ProcessPoolExecutor")
        retryer = Retrying(reraise=True, stop=stop_after_attempt(retries + 1))
        retrying_function = partial(retryer, function)

    def create_futures_func(input, **kwargs):
        return [
            (
                i,
                asyncio.wrap_future(
                    concurrent_executor.submit(retrying_function, i, **kwargs)
                ),
            )
            for i in input
        ]

    def create_futures_func_multiprocessing(input, **kwargs):
        # Pickle the function, args, and kwargs using cloudpickle.
        # They will be unpickled by unpickle_and_call.
        pickled_kwargs = {k: cloudpickle.dumps(v) for k, v in kwargs.items()}
        return [
            (
                i,
                asyncio.wrap_future(
                    concurrent_executor.submit(
                        unpickle_and_call,
                        cloudpickle.dumps(retrying_function),
                        cloudpickle.dumps(i),
                        **pickled_kwargs,
                    )
                ),
            )
            for i in input
        ]

    if isinstance(concurrent_executor, ProcessPoolExecutor):
        create_futures = create_futures_func_multiprocessing
    else:
        create_futures = create_futures_func
    async for result in async_map_unordered(
        create_futures,
        input,
        use_backups=use_backups,
        batch_size=batch_size,
        return_stats=return_stats,
        name=name,
        **kwargs,
    ):
        yield result


def pipeline_to_stream(
    concurrent_executor: Executor, name: str, pipeline: CubedPipeline, **kwargs
) -> Stream:
    return stream.iterate(
        map_unordered(
            concurrent_executor,
            run_func,
            pipeline.mappable,
            return_stats=True,
            name=name,
            func=pipeline.function,
            config=pipeline.config,
            **kwargs,
        )
    )


async def async_execute_dag(
    dag: MultiDiGraph,
    callbacks: Optional[Sequence[Callback]] = None,
    resume: Optional[bool] = None,
    spec: Optional[Spec] = None,
    compute_arrays_in_parallel: Optional[bool] = None,
    **kwargs,
) -> None:
    concurrent_executor: Executor
    use_processes = kwargs.pop("use_processes", False)
    if use_processes:
        max_workers = kwargs.pop("max_workers", None)
        context = multiprocessing.get_context("spawn")
        # max_tasks_per_child is only supported from Python 3.11
        concurrent_executor = ProcessPoolExecutor(
            max_workers=max_workers, mp_context=context, max_tasks_per_child=1
        )
    else:
        concurrent_executor = ThreadPoolExecutor()
    try:
        if not compute_arrays_in_parallel:
            # run one pipeline at a time
            for name, node in visit_nodes(dag, resume=resume):
                handle_operation_start_callbacks(callbacks, name)
                st = pipeline_to_stream(
                    concurrent_executor, name, node["pipeline"], **kwargs
                )
                async with st.stream() as streamer:
                    async for _, stats in streamer:
                        handle_callbacks(callbacks, stats)
        else:
            for gen in visit_node_generations(dag, resume=resume):
                # run pipelines in the same topological generation in parallel by merging their streams
                streams = [
                    pipeline_to_stream(
                        concurrent_executor, name, node["pipeline"], **kwargs
                    )
                    for name, node in gen
                ]
                merged_stream = stream.merge(*streams)
                async with merged_stream.stream() as streamer:
                    async for _, stats in streamer:
                        handle_callbacks(callbacks, stats)

    finally:
        # don't wait for any cancelled tasks
        concurrent_executor.shutdown(wait=False)


class AsyncPythonDagExecutor(DagExecutor):
    """An execution engine that uses Python asyncio."""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

        # Tell NumPy to use a single thread
        # from https://stackoverflow.com/questions/30791550/limit-number-of-threads-in-numpy
        os.environ["MKL_NUM_THREADS"] = "1"
        os.environ["NUMEXPR_NUM_THREADS"] = "1"
        os.environ["OMP_NUM_THREADS"] = "1"
        os.environ["VECLIB_MAXIMUM_THREADS"] = "1"

    @property
    def name(self) -> str:
        return "processes" if self.kwargs.get("use_processes", False) else "threads"

    def execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        resume: Optional[bool] = None,
        spec: Optional[Spec] = None,
        compute_id: Optional[str] = None,
        **kwargs,
    ) -> None:
        merged_kwargs = {**self.kwargs, **kwargs}
        asyncio.run(
            async_execute_dag(
                dag,
                callbacks=callbacks,
                resume=resume,
                spec=spec,
                compute_id=compute_id,
                **merged_kwargs,
            )
        )



from typing import Any, Dict, Optional

import networkx as nx

from cubed.storage.zarr import open_if_lazy_zarr_array


def already_computed(
    name, dag, nodes: Dict[str, Any], resume: Optional[bool] = None
) -> bool:
    """
    Return True if the array for a node doesn't have a pipeline to compute it,
    or if all its outputs have already been computed (all chunks are present).
    """
    pipeline = nodes[name].get("pipeline", None)
    if pipeline is None:
        return True

    # if no outputs have targets then need to compute (this is the create-arrays case)
    if all(
        [nodes[output].get("target", None) is None for output in dag.successors(name)]
    ):
        return False

    if resume:
        for output in dag.successors(name):
            target = nodes[output].get("target", None)
            if target is not None:
                target = open_if_lazy_zarr_array(target)
                # this check can be expensive since it has to list the directory to find nchunks_initialized
                if target.ndim == 0 or target.nchunks_initialized != target.nchunks:
                    return False
        return True

    return False


def visit_nodes(dag, resume=None):
    """Return a generator that visits the nodes in the DAG in topological order."""
    nodes = {n: d for (n, d) in dag.nodes(data=True)}
    for name in list(nx.topological_sort(dag)):
        if already_computed(name, dag, nodes, resume=resume):
            continue
        yield name, nodes[name]


def visit_node_generations(dag, resume=None):
    """Return a generator that visits the nodes in the DAG in groups of topological generations."""
    nodes = {n: d for (n, d) in dag.nodes(data=True)}
    for names in nx.topological_generations(dag):
        gen = [
            (name, nodes[name])
            for name in names
            if not already_computed(name, dag, nodes, resume=resume)
        ]
        if len(gen) > 0:
            yield gen



from itertools import product
from math import prod

import numpy as np
import pytest

from cubed.storage.virtual import virtual_empty, virtual_offsets


@pytest.mark.parametrize(
    "shape,chunks,index",
    [
        ((3,), (2,), 2),
        ((3, 2), (2, 1), (2, 1)),
        ((3, 2), (2, 1), (2, slice(0, 1))),
        ((3, 2), (2, 1), (slice(1, 3), 1)),
        ((3, 2), (2, 1), (slice(1, 3), slice(0, 1))),
    ],
)
def test_virtual_empty(shape, chunks, index):
    # array contents can be any uninitialized values, so
    # just check shapes not values
    v_empty = virtual_empty(shape, dtype=np.int32, chunks=chunks)
    empty = np.empty(shape, dtype=np.int32)
    assert v_empty[index].shape == empty[index].shape
    assert v_empty[...].shape == empty[...].shape


@pytest.mark.parametrize("shape", [(), (3,), (3, 2)])
def test_virtual_offsets(shape):
    v_offsets = virtual_offsets(shape)
    offsets = np.arange(prod(shape)).reshape(shape, order="C")
    for t in product(*(range(n) for n in shape)):
        assert v_offsets[t] == offsets[t]

    # test some length 1 slices
    if len(shape) == 1:
        assert v_offsets[1:2] == offsets[1:2]
    elif len(shape) == 2:
        assert v_offsets[1:2, 0:1] == offsets[1:2, 0:1]


def test_virtual_offsets_fails():
    with pytest.raises(NotImplementedError):
        v_offsets = virtual_offsets((3,))
        v_offsets[0:2]

    with pytest.raises(NotImplementedError):
        v_offsets = virtual_offsets((3, 2))
        v_offsets[0:2, 1]



import math
import shutil
from functools import partial, reduce

import pytest

from cubed.core.ops import partial_reduce
from cubed.core.optimization import multiple_inputs_optimize_dag

pytest.importorskip("lithops")

import cubed
import cubed.array_api as xp
import cubed.random
from cubed.backend_array_api import namespace as nxp
from cubed.extensions.history import HistoryCallback
from cubed.runtime.executors.lithops import LithopsDagExecutor
from cubed.tests.utils import LITHOPS_LOCAL_CONFIG


@pytest.fixture()
def spec(tmp_path, reserved_mem):
    return cubed.Spec(tmp_path, allowed_mem=2_000_000_000, reserved_mem=reserved_mem)


@pytest.fixture(scope="module")
def reserved_mem():
    executor = LithopsDagExecutor(config=LITHOPS_LOCAL_CONFIG)
    res = cubed.measure_reserved_mem(executor) * 1.1  # add some wiggle room
    return round_up_to_multiple(res, 10_000_000)  # round up to nearest multiple of 10MB


def round_up_to_multiple(x, multiple=10):
    """Round up to the nearest multiple"""
    return math.ceil(x / multiple) * multiple


# Array Object


@pytest.mark.slow
def test_index(tmp_path, spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = a[1:, :]
    run_operation(tmp_path, "index", b)


@pytest.mark.slow
def test_index_step(tmp_path, spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = a[::2, :]
    run_operation(tmp_path, "index_step", b)


# Creation Functions


@pytest.mark.slow
def test_eye(tmp_path, spec):
    a = xp.eye(10000, 10000, chunks=(5000, 5000), spec=spec)
    run_operation(tmp_path, "eye", a)


@pytest.mark.slow
def test_tril(tmp_path, spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.tril(a)
    run_operation(tmp_path, "tril", b)


# Elementwise Functions


@pytest.mark.slow
def test_add(tmp_path, spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    c = xp.add(a, b)
    run_operation(tmp_path, "add", c)


@pytest.mark.slow
def test_add_reduce_left(tmp_path, spec):
    # Perform the `add` operation repeatedly on pairs of arrays, also known as fold left.
    # See https://en.wikipedia.org/wiki/Fold_(higher-order_function)
    #
    # o   o
    #  \ /
    #   o   o
    #    \ /
    #     o   o
    #      \ /
    #       o
    #
    # Fusing fold left operations will result in a single fused operation.
    n_arrays = 10
    arrs = [
        cubed.random.random((10000, 10000), chunks=(5000, 5000), spec=spec)
        for _ in range(n_arrays)
    ]
    result = reduce(lambda x, y: xp.add(x, y), arrs)
    opt_fn = partial(multiple_inputs_optimize_dag, max_total_source_arrays=n_arrays * 2)
    run_operation(tmp_path, "add_reduce_left", result, optimize_function=opt_fn)


@pytest.mark.slow
def test_add_reduce_right(tmp_path, spec):
    # Perform the `add` operation repeatedly on pairs of arrays, also known as fold right.
    # See https://en.wikipedia.org/wiki/Fold_(higher-order_function)
    #
    #     o   o
    #      \ /
    #   o   o
    #    \ /
    # o   o
    #  \ /
    #   o
    #
    # Note that fusing fold right operations will result in unbounded memory usage unless care
    # is taken to limit fusion - which `multiple_inputs_optimize_dag` will do, with the result
    # that there is more than one fused operation (not a single fused operation).
    n_arrays = 10
    arrs = [
        cubed.random.random((10000, 10000), chunks=(5000, 5000), spec=spec)
        for _ in range(n_arrays)
    ]
    result = reduce(lambda x, y: xp.add(y, x), reversed(arrs))
    opt_fn = partial(multiple_inputs_optimize_dag, max_total_source_arrays=n_arrays * 2)
    run_operation(tmp_path, "add_reduce_right", result, optimize_function=opt_fn)


@pytest.mark.slow
def test_negative(tmp_path, spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.negative(a)
    run_operation(tmp_path, "negative", b)


# Linear Algebra Functions


@pytest.mark.slow
def test_matmul(tmp_path, spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    c = xp.astype(a, xp.float32)
    d = xp.astype(b, xp.float32)
    e = xp.matmul(c, d)
    run_operation(tmp_path, "matmul", e)


@pytest.mark.slow
def test_matrix_transpose(tmp_path, spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.matrix_transpose(a)
    run_operation(tmp_path, "matrix_transpose", b)


@pytest.mark.slow
def test_tensordot(tmp_path, spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    c = xp.astype(a, xp.float32)
    d = xp.astype(b, xp.float32)
    e = xp.tensordot(c, d, axes=1)
    run_operation(tmp_path, "tensordot", e)


# Manipulation Functions


@pytest.mark.slow
def test_concat(tmp_path, spec):
    # Note 'a' has one fewer element in axis=0 to force chunking to cross array boundaries
    a = cubed.random.random(
        (9999, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    c = xp.concat((a, b), axis=0)
    run_operation(tmp_path, "concat", c)


@pytest.mark.slow
def test_reshape(tmp_path, spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    # need intermediate reshape due to limitations in Dask's reshape_rechunk
    b = xp.reshape(a, (5000, 2, 10000))
    c = xp.reshape(b, (5000, 20000))
    run_operation(tmp_path, "reshape", c)


@pytest.mark.slow
def test_stack(tmp_path, spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    c = xp.stack((a, b), axis=0)
    run_operation(tmp_path, "stack", c)


# Searching Functions


@pytest.mark.slow
def test_argmax(tmp_path, spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.argmax(a, axis=0)
    run_operation(tmp_path, "argmax", b)


# Statistical Functions


@pytest.mark.slow
def test_max(tmp_path, spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.max(a, axis=0)
    run_operation(tmp_path, "max", b)


@pytest.mark.slow
def test_mean(tmp_path, spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.mean(a, axis=0)
    run_operation(tmp_path, "mean", b)


@pytest.mark.slow
def test_sum_partial_reduce(tmp_path, spec):
    a = cubed.random.random(
        (40000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = partial_reduce(a, nxp.sum, split_every={0: 8})
    run_operation(tmp_path, "sum_partial_reduce", b)


# Internal functions


def run_operation(tmp_path, name, result_array, *, optimize_function=None):
    # result_array.visualize(f"cubed-{name}-unoptimized", optimize_graph=False)
    # result_array.visualize(f"cubed-{name}", optimize_function=optimize_function)
    executor = LithopsDagExecutor(config=LITHOPS_LOCAL_CONFIG)
    hist = HistoryCallback()
    # use store=None to write to temporary zarr
    cubed.to_zarr(
        result_array,
        store=None,
        executor=executor,
        callbacks=[hist],
        optimize_function=optimize_function,
    )

    df = hist.stats_df
    print(df)

    # check projected_mem_utilization does not exceed 1
    assert (df["projected_mem_utilization"] <= 1.0).all()

    # delete temp files for this test immediately since they are so large
    shutil.rmtree(tmp_path)



import platform
import random
from functools import partial

import dill
import numpy as np
import pytest
import zarr
from numpy.testing import assert_array_equal

import cubed
import cubed.array_api as xp
import cubed.random
from cubed.backend_array_api import namespace as nxp
from cubed.core.ops import merge_chunks, partial_reduce, tree_reduce
from cubed.core.optimization import fuse_all_optimize_dag, multiple_inputs_optimize_dag
from cubed.tests.utils import (
    ALL_EXECUTORS,
    MAIN_EXECUTORS,
    MODAL_EXECUTORS,
    TaskCounter,
    create_zarr,
)


@pytest.fixture()
def spec(tmp_path):
    return cubed.Spec(tmp_path, allowed_mem=100000)


@pytest.fixture(
    scope="module",
    params=MAIN_EXECUTORS,
    ids=[executor.name for executor in MAIN_EXECUTORS],
)
def executor(request):
    return request.param


@pytest.fixture(
    scope="module",
    params=ALL_EXECUTORS,
    ids=[executor.name for executor in ALL_EXECUTORS],
)
def any_executor(request):
    return request.param


@pytest.fixture(
    scope="module",
    params=MODAL_EXECUTORS,
    ids=[executor.name for executor in MODAL_EXECUTORS],
)
def modal_executor(request):
    return request.param


def test_as_array_fails(spec):
    a = np.ones((1000, 1000))
    with pytest.raises(
        ValueError,
        match="Size of in memory array is 8.0 MB which exceeds maximum of 1.0 MB.",
    ):
        xp.asarray(a, chunks=(100, 100), spec=spec)


def test_regular_chunks(spec):
    xp.ones((5, 5), chunks=((2, 2, 1), (5,)), spec=spec)
    with pytest.raises(ValueError):
        xp.ones((5, 5), chunks=((2, 1, 2), (5,)), spec=spec)


class WrappedArray:
    def __init__(self, x):
        self.x = x
        self.dtype = x.dtype
        self.shape = x.shape
        self.ndim = len(x.shape)

    def __array__(self, dtype=None):
        return np.asarray(self.x, dtype=dtype)

    def __getitem__(self, i):
        return WrappedArray(self.x[i])


@pytest.mark.parametrize(
    "x,chunks,asarray",
    [
        (np.arange(25).reshape((5, 5)), (5, 5), None),
        (np.arange(25).reshape((5, 5)), (3, 2), True),
        (np.arange(25).reshape((5, 5)), -1, True),
        (np.array([[1]]), 1, None),
    ],
)
def test_from_array(x, chunks, asarray):
    a = cubed.from_array(WrappedArray(x), chunks=chunks, asarray=asarray)
    assert isinstance(a, cubed.Array)
    assert_array_equal(a, x)


def test_from_array_zarr(tmp_path, spec):
    store = store = tmp_path / "source.zarr"
    za = create_zarr(
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        chunks=(2, 2),
        store=store,
    )
    a = cubed.from_array(za, spec=spec)
    assert_array_equal(a, za)


@pytest.mark.parametrize("path", [None, "sub", "sub/group"])
def test_from_zarr(tmp_path, spec, executor, path):
    store = store = tmp_path / "source.zarr"
    create_zarr(
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        chunks=(2, 2),
        store=store,
        path=path,
    )
    a = cubed.from_zarr(store, path=path, spec=spec)
    assert_array_equal(
        a.compute(executor=executor), np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    )


def test_store(tmp_path, spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)

    store = tmp_path / "source.zarr"
    target = zarr.empty(a.shape, store=store)

    cubed.store(a, target)
    assert_array_equal(target, np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))


def test_store_multiple(tmp_path, spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)

    store1 = tmp_path / "source1.zarr"
    target1 = zarr.empty(a.shape, store=store1)
    store2 = tmp_path / "source2.zarr"
    target2 = zarr.empty(b.shape, store=store2)

    cubed.store([a, b], [target1, target2])
    assert_array_equal(target1, np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))
    assert_array_equal(target2, np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]]))


def test_store_fails(tmp_path, spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    store = tmp_path / "source.zarr"
    target = zarr.empty(a.shape, store=store)

    with pytest.raises(
        ValueError, match=r"Different number of sources \(2\) and targets \(1\)"
    ):
        cubed.store([a, b], [target])

    with pytest.raises(ValueError, match="All sources must be cubed array objects"):
        cubed.store([1], [target])


@pytest.mark.parametrize("path", [None, "sub", "sub/group"])
def test_to_zarr(tmp_path, spec, executor, path):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    store = zarr.storage.DirectoryStore(tmp_path / "output.zarr")
    cubed.to_zarr(a, store, path=path, executor=executor)
    res = zarr.open_array(store, path=path)
    assert_array_equal(res[:], np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))


def test_map_blocks_with_kwargs(spec, executor):
    # based on dask test
    a = xp.asarray([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], chunks=5, spec=spec)
    b = cubed.map_blocks(nxp.max, a, axis=0, keepdims=True, dtype=a.dtype, chunks=(1,))
    assert_array_equal(b.compute(executor=executor), np.array([4, 9]))


def test_map_blocks_with_block_id(spec, executor):
    # based on dask test
    def func(block, block_id=None, c=0):
        return nxp.ones_like(block) * int(sum(block_id)) + c

    a = xp.arange(10, dtype="int64", chunks=(2,))
    b = cubed.map_blocks(func, a, dtype="int64")

    assert_array_equal(
        b.compute(executor=executor),
        np.array([0, 0, 1, 1, 2, 2, 3, 3, 4, 4], dtype="int64"),
    )

    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = cubed.map_blocks(func, a, dtype="int64")

    assert_array_equal(
        b.compute(executor=executor),
        np.array([[0, 0, 1], [0, 0, 1], [1, 1, 2]], dtype="int64"),
    )

    c = cubed.map_blocks(func, a, dtype="int64", c=1)

    assert_array_equal(
        c.compute(executor=executor),
        np.array([[0, 0, 1], [0, 0, 1], [1, 1, 2]], dtype="int64") + 1,
    )


def test_map_blocks_no_array_args(spec, executor):
    def func(block, block_id=None):
        return nxp.ones_like(block) * int(sum(block_id))

    a = cubed.map_blocks(func, dtype="int64", chunks=((5, 3),), spec=spec)
    assert a.chunks == ((5, 3),)

    assert_array_equal(
        a.compute(executor=executor),
        np.array([0, 0, 0, 0, 0, 1, 1, 1], dtype="int64"),
    )


def test_map_blocks_with_different_block_shapes(spec):
    def func(x, y):
        return x

    a = xp.asarray([[[12, 13]]], spec=spec)
    b = xp.asarray([14, 15], spec=spec)
    c = cubed.map_blocks(
        func, a, b, dtype="int64", chunks=(1, 1, 2), drop_axis=2, new_axis=2
    )
    assert_array_equal(c.compute(), np.array([[[12, 13]]]))


def test_multiple_ops(spec, executor):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    d = xp.negative(c)
    assert_array_equal(
        d.compute(executor=executor),
        np.array([[-2, -3, -4], [-5, -6, -7], [-8, -9, -10]]),
    )


@pytest.mark.parametrize("new_chunks", [(1, 2), {0: 1, 1: 2}])
def test_rechunk(spec, executor, new_chunks):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 1), spec=spec)
    b = a.rechunk(new_chunks)
    assert_array_equal(
        b.compute(executor=executor),
        np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
    )


def test_rechunk_same_chunks(spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 1), spec=spec)
    b = a.rechunk((2, 1))
    task_counter = TaskCounter()
    res = b.compute(callbacks=[task_counter])
    # no tasks should have run since chunks are same
    assert task_counter.value == 0

    assert_array_equal(res, np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))


# see also test_rechunk.py
def test_rechunk_intermediate(tmp_path):
    spec = cubed.Spec(tmp_path, allowed_mem=4 * 8 * 4)
    a = xp.ones((4, 4), chunks=(1, 4), spec=spec)
    b = a.rechunk((4, 1))
    assert_array_equal(b.compute(), np.ones((4, 4)))
    intermediates = [n for (n, d) in b.plan.dag.nodes(data=True) if "-int" in d["name"]]
    assert len(intermediates) == 1


def test_compute_is_idempotent(spec, executor):
    a = xp.ones((3, 3), chunks=(2, 2), spec=spec)
    b = xp.negative(a)
    assert_array_equal(b.compute(executor=executor), -np.ones((3, 3)))
    assert_array_equal(b.compute(executor=executor), -np.ones((3, 3)))


def test_default_spec(executor):
    # default spec works for small computations
    a = xp.ones((3, 3), chunks=(2, 2))
    b = xp.negative(a)
    assert_array_equal(
        b.compute(executor=executor),
        -np.ones((3, 3)),
    )


def test_default_spec_allowed_mem_exceeded():
    # default spec fails for large computations
    a = xp.ones((20000, 1000), chunks=(10000, 1000))
    with pytest.raises(ValueError):
        xp.negative(a)


def test_default_spec_config_override():
    # override default spec to increase allowed_mem
    from cubed import config

    with config.set({"spec.allowed_mem": "500MB"}):
        a = xp.ones((20000, 1000), chunks=(10000, 1000))
        b = xp.negative(a)
        assert_array_equal(b.compute(), -np.ones((20000, 1000)))


def test_different_specs(tmp_path):
    spec1 = cubed.Spec(tmp_path, allowed_mem=100000)
    spec2 = cubed.Spec(tmp_path, allowed_mem=200000)
    a = xp.ones((3, 3), chunks=(2, 2), spec=spec1)
    b = xp.ones((3, 3), chunks=(2, 2), spec=spec2)
    with pytest.raises(ValueError):
        xp.add(a, b)


@pytest.mark.parametrize(
    "input_value, expected_value",
    [
        (500, 500),
        (100_000, 100_000),
        (50.0, 50),
        ("500B", 500),
        ("1kB", 1000),
        ("1MB", 1000**2),
        ("1GB", 1000**3),
        ("1TB", 1000**4),
        ("1PB", 1000**5),
        ("100_000", 100_000),
        ("1.2MB", 1.2 * 1000**2),
        ("1 MB", 1000**2),
        ("1.2 MB", 1.2 * 1000**2),
    ],
)
def test_convert_to_bytes(input_value, expected_value):
    spec = cubed.Spec(allowed_mem=input_value)
    assert spec.allowed_mem == expected_value


@pytest.mark.parametrize(
    "input_value",
    [
        "1EB",  # EB is not a valid unit in this function
        "1kb",  # lower-case k is not valid
        "invalid",  # completely invalid input
        -512,  # negative integer
        "kB",  # only unit, no value
        "1.1B",  # can't have a fractional number of bytes
    ],
)
def test_convert_to_bytes_error(input_value):
    with pytest.raises(ValueError):
        cubed.Spec(allowed_mem=input_value)


def test_reduction_multiple_rounds(tmp_path, executor):
    spec = cubed.Spec(tmp_path, allowed_mem=1000)
    a = xp.ones((100, 10), dtype=np.uint8, chunks=(1, 10), spec=spec)
    b = xp.sum(a, axis=0, dtype=np.uint8)
    # check that there is > 1 blockwise step (after optimization)
    blockwises = [
        n
        for (n, d) in b.plan.dag.nodes(data=True)
        if d.get("op_name", None) == "blockwise"
    ]
    assert len(blockwises) > 1
    assert b.plan.max_projected_mem() <= 1000
    assert_array_equal(b.compute(executor=executor), np.ones((100, 10)).sum(axis=0))


def test_reduction_not_enough_memory(tmp_path):
    spec = cubed.Spec(tmp_path, allowed_mem=50)
    a = xp.ones((100, 10), dtype=np.uint8, chunks=(1, 10), spec=spec)
    with pytest.raises(ValueError, match=r"Not enough memory for reduction"):
        # only a problem with the old implementation, so set use_new_impl=False
        xp.sum(a, axis=0, dtype=np.uint8, use_new_impl=False)


def test_partial_reduce(spec):
    a = xp.asarray(np.arange(242).reshape((11, 22)), chunks=(3, 4), spec=spec)
    b = partial_reduce(a, np.sum, split_every={0: 2})
    c = partial_reduce(b, np.sum, split_every={0: 2})
    assert_array_equal(
        c.compute(), np.arange(242).reshape((11, 22)).sum(axis=0, keepdims=True)
    )


def test_tree_reduce(spec):
    a = xp.asarray(np.arange(242).reshape((11, 22)), chunks=(3, 4), spec=spec)
    b = tree_reduce(a, np.sum, axis=0, dtype=np.int64, split_every={0: 2})
    assert_array_equal(
        b.compute(), np.arange(242).reshape((11, 22)).sum(axis=0, keepdims=True)
    )


@pytest.mark.parametrize(
    "target_chunks, expected_chunksize",
    [
        ((2, 3), None),
        ((4, 3), None),
        ((2, 6), None),
        ((4, 6), None),
        ((12, 12), (10, 10)),
    ],
)
def test_merge_chunks(spec, target_chunks, expected_chunksize):
    a = xp.ones((10, 10), dtype=np.uint8, chunks=(2, 3), spec=spec)
    b = merge_chunks(a, target_chunks)
    assert b.chunksize == (expected_chunksize or target_chunks)
    assert_array_equal(b.compute(), np.ones((10, 10)))


@pytest.mark.parametrize(
    "target_chunks", [(2,), (2, 3, 1), (3, 2), (1, 3), (5, 5), (10, 10)]
)
def test_merge_chunks_fails(spec, target_chunks):
    a = xp.ones((10, 10), dtype=np.uint8, chunks=(2, 3), spec=spec)
    with pytest.raises(ValueError):
        merge_chunks(a, target_chunks)


def test_compute_multiple():
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2))
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2))
    c = xp.add(a, b)
    d = c * 2
    e = c * 3

    f = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2))
    g = f * 4

    dc, ec, gc = cubed.compute(d, e, g)

    an = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    bn = np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]])
    cn = an + bn
    dn = cn * 2
    en = cn * 3

    fn = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    gn = fn * 4

    assert_array_equal(dc, dn)
    assert_array_equal(ec, en)
    assert_array_equal(gc, gn)


def test_compute_multiple_different_specs(tmp_path):
    spec1 = cubed.Spec(tmp_path, allowed_mem=100000)
    spec2 = cubed.Spec(tmp_path, allowed_mem=200000)

    a1 = xp.ones((3, 3), chunks=(2, 2), spec=spec1)
    b1 = xp.ones((3, 3), chunks=(2, 2), spec=spec1)
    c1 = xp.add(a1, b1)

    a2 = xp.ones((3, 3), chunks=(2, 2), spec=spec2)
    b2 = xp.ones((3, 3), chunks=(2, 2), spec=spec2)
    c2 = xp.add(a2, b2)

    with pytest.raises(ValueError):
        cubed.compute(c1, c2)


def test_visualize(tmp_path):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=xp.float64, chunks=(2, 2))
    b = cubed.random.random((3, 3), chunks=(2, 2))
    c = xp.add(a, b)
    d = c.rechunk((3, 1))
    e = c * 3

    f = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2))
    g = f * 4

    assert not (tmp_path / "e.dot").exists()
    assert not (tmp_path / "e.png").exists()
    assert not (tmp_path / "e.svg").exists()
    assert not (tmp_path / "dg.svg").exists()

    e.visualize(filename=tmp_path / "e")
    assert (tmp_path / "e.svg").exists()

    e.visualize(filename=tmp_path / "e-hidden", show_hidden=True)
    assert (tmp_path / "e-hidden.svg").exists()

    e.visualize(filename=tmp_path / "e", format="png")
    assert (tmp_path / "e.png").exists()

    e.visualize(filename=tmp_path / "e", format="dot")
    assert (tmp_path / "e.dot").exists()

    # multiple arrays
    cubed.visualize(d, g, filename=tmp_path / "dg")
    assert (tmp_path / "dg.svg").exists()


def test_array_pickle(spec, executor):
    a = xp.asarray(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        chunks=(2, 2),
        spec=spec,
    )
    b = xp.asarray(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        chunks=(2, 2),
        spec=spec,
    )
    c = xp.matmul(a, b)

    # we haven't computed c yet, so pickle and unpickle, and check it still works
    # note we have to use dill which can serialize local functions, unlike pickle
    c = dill.loads(dill.dumps(c))

    x = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])
    y = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])
    expected = np.matmul(x, y)
    assert_array_equal(c.compute(executor=executor), expected)


@pytest.mark.skipif(platform.system() == "Windows", reason="does not run on windows")
def test_measure_reserved_mem(executor):
    pytest.importorskip("lithops")

    if executor.name != "lithops":
        pytest.skip(f"{executor.name} executor does not support measure_reserved_mem")

    reserved_memory = cubed.measure_reserved_mem(executor=executor)
    assert reserved_memory > 1_000_000  # over 1MB


# Test we can create a plan for arrays of up to 5PB, and 100s of billions of tasks
@pytest.mark.parametrize("factor", [10, 20, 50, 100, 200, 500, 1000, 2000, 5000])
def test_plan_scaling(tmp_path, factor):
    spec = cubed.Spec(tmp_path, allowed_mem="2GB")
    chunksize = 5000
    a = cubed.random.random(
        (factor * chunksize, factor * chunksize), chunks=chunksize, spec=spec
    )
    b = cubed.random.random(
        (factor * chunksize, factor * chunksize), chunks=chunksize, spec=spec
    )
    c = xp.matmul(a, b)

    assert c.plan.num_tasks() > 0
    c.visualize(filename=tmp_path / "c")


@pytest.mark.parametrize("t_length", [50, 500, 5000, 50000])
def test_plan_quad_means(tmp_path, t_length):
    # based on sizes from https://gist.github.com/TomNicholas/c6a28f7c22c6981f75bce280d3e28283
    spec = cubed.Spec(tmp_path, allowed_mem="2GB", reserved_mem="100MB")
    u = cubed.random.random((t_length, 1, 987, 1920), chunks=(10, 1, -1, -1), spec=spec)
    v = cubed.random.random((t_length, 1, 987, 1920), chunks=(10, 1, -1, -1), spec=spec)
    uv = u * v
    m = xp.mean(uv, axis=0, split_every=10, use_new_impl=True)

    assert m.plan.num_tasks() > 0
    m.visualize(
        filename=tmp_path / "quad_means_unoptimized",
        optimize_graph=False,
        show_hidden=True,
    )
    m.visualize(
        filename=tmp_path / "quad_means",
        optimize_function=multiple_inputs_optimize_dag,
        show_hidden=True,
    )


def quad_means(tmp_path, t_length):
    # based on sizes from https://gist.github.com/TomNicholas/c6a28f7c22c6981f75bce280d3e28283
    spec = cubed.Spec(tmp_path, allowed_mem="2GB", reserved_mem="100MB")
    u = cubed.random.random((t_length, 1, 987, 1920), chunks=(10, 1, -1, -1), spec=spec)
    v = cubed.random.random((t_length, 1, 987, 1920), chunks=(10, 1, -1, -1), spec=spec)
    uv = u * v
    m = xp.mean(uv, axis=0)
    return m


def test_quad_means(tmp_path, t_length=50):
    # run twice, with and without optimization
    # set the random seed to ensure deterministic results
    random.seed(42)
    m0 = quad_means(tmp_path, t_length)

    random.seed(42)
    m1 = quad_means(tmp_path, t_length)

    m1.visualize(
        filename=tmp_path / "quad_means", optimize_function=fuse_all_optimize_dag
    )

    cubed.to_zarr(m0, store=tmp_path / "result0")
    cubed.to_zarr(
        m1, store=tmp_path / "result1", optimize_function=fuse_all_optimize_dag
    )

    res0 = zarr.open_array(tmp_path / "result0")
    res1 = zarr.open_array(tmp_path / "result1")

    assert_array_equal(res0[:], res1[:])


def test_quad_means_zarr(tmp_path, t_length=50):
    # write inputs to Zarr first to test more realistic usage pattern
    spec = cubed.Spec(tmp_path, allowed_mem="2GB", reserved_mem="100MB")
    u = cubed.random.random((t_length, 1, 987, 1920), chunks=(10, 1, -1, -1), spec=spec)
    v = cubed.random.random((t_length, 1, 987, 1920), chunks=(10, 1, -1, -1), spec=spec)

    arrays = [u, v]
    paths = [f"{tmp_path}/u_{t_length}.zarr", f"{tmp_path}/v_{t_length}.zarr"]
    cubed.store(arrays, paths)

    u = cubed.from_zarr(f"{tmp_path}/u_{t_length}.zarr", spec=spec)
    v = cubed.from_zarr(f"{tmp_path}/v_{t_length}.zarr", spec=spec)
    uv = u * v
    m = xp.mean(uv, axis=0, use_new_impl=True, split_every=10)

    opt_fn = partial(multiple_inputs_optimize_dag, max_total_num_input_blocks=40)

    m.visualize(filename=tmp_path / "quad_means", optimize_function=opt_fn)

    cubed.to_zarr(m, store=tmp_path / "result", optimize_function=opt_fn)



import asyncio
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from aiostream import stream
from aiostream.core import Stream
from dask.distributed import Client
from networkx import MultiDiGraph

from cubed.runtime.executors.asyncio import async_map_unordered
from cubed.runtime.pipeline import visit_node_generations, visit_nodes
from cubed.runtime.types import Callback, CubedPipeline, DagExecutor
from cubed.runtime.utils import (
    execution_stats,
    gensym,
    handle_callbacks,
    handle_operation_start_callbacks,
)
from cubed.spec import Spec


# note we can't call `pipeline_func` just `func` here as it clashes with `dask.distributed.Client.map``
@execution_stats
def run_func(input, pipeline_func=None, config=None, name=None, compute_id=None):
    result = pipeline_func(input, config=config)
    return result


async def map_unordered(
    client: Client,
    map_function: Callable[..., Any],
    map_iterdata: Iterable[Union[List[Any], Tuple[Any, ...], Dict[str, Any]]],
    retries: int = 2,
    use_backups: bool = True,
    batch_size: Optional[int] = None,
    return_stats: bool = False,
    name: Optional[str] = None,
    **kwargs,
) -> AsyncIterator[Any]:
    def create_futures_func(input, **kwargs):
        input = list(input)  # dask expects a sequence (it calls `len` on it)
        key = name or gensym("map")
        key = key.replace("-", "_")  # otherwise array number is not shown on dashboard
        return [
            (i, asyncio.ensure_future(f))
            for i, f in zip(
                input,
                client.map(map_function, input, key=key, retries=retries, **kwargs),
            )
        ]

    def create_backup_futures_func(input, **kwargs):
        input = list(input)  # dask expects a sequence (it calls `len` on it)
        key = name or gensym("backup")
        key = key.replace("-", "_")  # otherwise array number is not shown on dashboard
        return [
            (i, asyncio.ensure_future(f))
            for i, f in zip(input, client.map(map_function, input, key=key, **kwargs))
        ]

    async for result in async_map_unordered(
        create_futures_func,
        map_iterdata,
        use_backups=use_backups,
        create_backup_futures_func=create_backup_futures_func,
        batch_size=batch_size,
        return_stats=return_stats,
        name=name,
        **kwargs,
    ):
        yield result


def pipeline_to_stream(
    client: Client, name: str, pipeline: CubedPipeline, **kwargs
) -> Stream:
    return stream.iterate(
        map_unordered(
            client,
            run_func,
            pipeline.mappable,
            return_stats=True,
            name=name,
            pipeline_func=pipeline.function,
            config=pipeline.config,
            **kwargs,
        )
    )


def check_runtime_memory(spec, client):
    allowed_mem = spec.allowed_mem if spec is not None else None
    scheduler_info = client.scheduler_info()
    workers = scheduler_info.get("workers")
    if workers is None or len(workers) == 0:
        raise ValueError("Cluster has no workers running")
    runtime_memory = min(w["memory_limit"] // w["nthreads"] for w in workers.values())
    if allowed_mem is not None:
        if runtime_memory < allowed_mem:
            raise ValueError(
                f"Runtime memory ({runtime_memory}) is less than allowed_mem ({allowed_mem})"
            )


async def async_execute_dag(
    dag: MultiDiGraph,
    callbacks: Optional[Sequence[Callback]] = None,
    resume: Optional[bool] = None,
    spec: Optional[Spec] = None,
    compute_arrays_in_parallel: Optional[bool] = None,
    compute_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> None:
    compute_kwargs = compute_kwargs or {}
    async with Client(asynchronous=True, **compute_kwargs) as client:
        if spec is not None:
            check_runtime_memory(spec, client)
        if not compute_arrays_in_parallel:
            # run one pipeline at a time
            for name, node in visit_nodes(dag, resume=resume):
                handle_operation_start_callbacks(callbacks, name)
                st = pipeline_to_stream(client, name, node["pipeline"], **kwargs)
                async with st.stream() as streamer:
                    async for _, stats in streamer:
                        handle_callbacks(callbacks, stats)
        else:
            for gen in visit_node_generations(dag, resume=resume):
                # run pipelines in the same topological generation in parallel by merging their streams
                streams = [
                    pipeline_to_stream(client, name, node["pipeline"], **kwargs)
                    for name, node in gen
                ]
                merged_stream = stream.merge(*streams)
                async with merged_stream.stream() as streamer:
                    async for _, stats in streamer:
                        handle_callbacks(callbacks, stats)


class AsyncDaskDistributedExecutor(DagExecutor):
    """An execution engine that uses Dask Distributed's async API."""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @property
    def name(self) -> str:
        return "dask"

    def execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        resume: Optional[bool] = None,
        spec: Optional[Spec] = None,
        compute_id: Optional[str] = None,
        compute_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        merged_kwargs = {**self.kwargs, **kwargs}
        asyncio.run(
            async_execute_dag(
                dag,
                callbacks=callbacks,
                resume=resume,
                spec=spec,
                compute_kwargs=compute_kwargs,
                compute_id=compute_id,
                **merged_kwargs,
            )
        )



import os

import modal

from cubed.runtime.utils import execute_with_stats

RUNTIME_MEMORY_MIB = 2000

stub = modal.Stub("cubed-stub")

requirements_file = os.getenv("CUBED_MODAL_REQUIREMENTS_FILE")

if requirements_file:
    image = modal.Image.debian_slim().pip_install_from_requirements(requirements_file)
    aws_image = image
    gcp_image = image
else:
    aws_image = modal.Image.debian_slim().pip_install(
        [
            "array-api-compat",
            "donfig",
            "fsspec",
            "mypy_extensions",  # for rechunker
            "networkx",
            "pytest-mock",  # TODO: only needed for tests
            "s3fs",
            "tenacity",
            "toolz",
            "zarr",
        ]
    )
    gcp_image = modal.Image.debian_slim().pip_install(
        [
            "array-api-compat",
            "donfig",
            "fsspec",
            "mypy_extensions",  # for rechunker
            "networkx",
            "pytest-mock",  # TODO: only needed for tests
            "gcsfs",
            "tenacity",
            "toolz",
            "zarr",
        ]
    )


def check_runtime_memory(spec):
    allowed_mem = spec.allowed_mem if spec is not None else None
    runtime_memory = RUNTIME_MEMORY_MIB * 1024 * 1024
    if allowed_mem is not None:
        if runtime_memory < allowed_mem:
            raise ValueError(
                f"Runtime memory ({runtime_memory}) is less than allowed_mem ({allowed_mem})"
            )


@stub.function(
    image=aws_image,
    secrets=[modal.Secret.from_name("my-aws-secret")],
    memory=RUNTIME_MEMORY_MIB,
    retries=2,
    cloud="aws",
)
def run_remotely(input, func=None, config=None, name=None, compute_id=None):
    print(f"running remotely on {input} in {os.getenv('MODAL_REGION')}")
    # note we can't use the execution_stat decorator since it doesn't work with modal decorators
    result, stats = execute_with_stats(func, input, config=config)
    return result, stats


# For GCP we need to use a class so we can set up credentials by hooking into the container lifecycle
@stub.cls(
    image=gcp_image,
    secrets=[modal.Secret.from_name("my-googlecloud-secret")],
    memory=RUNTIME_MEMORY_MIB,
    retries=2,
    cloud="gcp",
)
class Container:
    @modal.enter()
    def set_up_credentials(self):
        json = os.environ["SERVICE_ACCOUNT_JSON"]
        path = os.path.abspath("application_credentials.json")
        with open(path, "w") as f:
            f.write(json)
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = path

    @modal.method()
    def run_remotely(self, input, func=None, config=None, name=None, compute_id=None):
        print(f"running remotely on {input} in {os.getenv('MODAL_REGION')}")
        # note we can't use the execution_stat decorator since it doesn't work with modal decorators
        result, stats = execute_with_stats(func, input, config=config)
        return result, stats



import time
from functools import partial
from itertools import islice

from cubed.runtime.types import OperationStartEvent, TaskEndEvent
from cubed.utils import peak_measured_mem

sym_counter = 0


def gensym(name: str) -> str:
    global sym_counter
    sym_counter += 1
    return f"{name}-{sym_counter:03}"


def execute_with_stats(function, *args, **kwargs):
    """Invoke function and measure timing information and peak memory usage.

    Returns the result of the function call and a stats dictionary.
    """

    peak_measured_mem_start = peak_measured_mem()
    function_start_tstamp = time.time()
    result = function(*args, **kwargs)
    function_end_tstamp = time.time()
    peak_measured_mem_end = peak_measured_mem()
    return result, dict(
        function_start_tstamp=function_start_tstamp,
        function_end_tstamp=function_end_tstamp,
        peak_measured_mem_start=peak_measured_mem_start,
        peak_measured_mem_end=peak_measured_mem_end,
    )


def execution_stats(func):
    """Decorator to measure timing information and peak memory usage of a function call."""

    return partial(execute_with_stats, func)


def handle_operation_start_callbacks(callbacks, name):
    if callbacks is not None:
        event = OperationStartEvent(name)
        [callback.on_operation_start(event) for callback in callbacks]


def handle_callbacks(callbacks, stats):
    """Construct a TaskEndEvent from stats and send to all callbacks."""

    if callbacks is not None:
        if "task_result_tstamp" not in stats:
            task_result_tstamp = time.time()
            event = TaskEndEvent(
                task_result_tstamp=task_result_tstamp,
                **stats,
            )
        else:
            event = TaskEndEvent(**stats)
        [callback.on_task_end(event) for callback in callbacks]


# this will be in Python 3.12 https://docs.python.org/3.12/library/itertools.html#itertools.batched
def batched(iterable, n):
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch



from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

from lithops import FunctionExecutor
from lithops.future import ResponseFuture
from lithops.wait import (
    ALL_COMPLETED,
    ALWAYS,
    ANY_COMPLETED,
    THREADPOOL_SIZE,
    WAIT_DUR_SEC,
)
from six import reraise


class RetryingFuture:
    """
    A wrapper around Lithops `ResponseFuture` that takes care of retries.
    """

    def __init__(
        self,
        response_future: ResponseFuture,
        map_function: Callable[..., Any],
        input: Any,
        map_kwargs: Any = None,
        retries: Optional[int] = None,
        group_name: Optional[str] = None,
    ):
        self.response_future = response_future
        self.map_function = map_function
        self.input = input
        self.map_kwargs = map_kwargs or {}
        self.retries = retries or 0
        self.group_name = group_name
        self.failure_count = 0
        self.cancelled = False

    def _inc_failure_count(self):
        self.failure_count += 1

    def _should_retry(self):
        return not self.cancelled and self.failure_count <= self.retries

    def _retry(self, function_executor: FunctionExecutor):
        inputs = [self.input]
        futures_list = function_executor.map(
            self.map_function, inputs, **self.map_kwargs
        )
        self.response_future = futures_list[0]

    def cancel(self):
        # cancelling will prevent any further retries, but won't affect any running tasks
        self.cancelled = True

    @property
    def done(self):
        return self.response_future.done

    @property
    def error(self):
        return self.response_future.error

    @property
    def _exception(self):
        return self.response_future._exception

    @property
    def stats(self):
        return self.response_future.stats

    def status(
        self,
        throw_except: bool = True,
        internal_storage: Any = None,
        check_only: bool = False,
    ):
        stat = self.response_future.status(
            throw_except=throw_except,
            internal_storage=internal_storage,
            check_only=check_only,
        )
        if self.response_future.error:
            reraise(*self.response_future._exception)
        return stat

    def result(self, throw_except: bool = True, internal_storage: Any = None):
        res = self.response_future.result(
            throw_except=throw_except, internal_storage=internal_storage
        )
        if self.response_future.error:
            reraise(*self.response_future._exception)
        return res


class RetryingFunctionExecutor:
    """
    A wrapper around Lithops `FunctionExecutor` that supports retries.
    """

    def __init__(self, executor: FunctionExecutor):
        self.executor = executor

    def __enter__(self):
        self.executor.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.executor.__exit__(exc_type, exc_value, traceback)

    def map(
        self,
        map_function: Callable[..., Any],
        map_iterdata: Iterable[Union[List[Any], Tuple[Any, ...], Dict[str, Any]]],
        *,
        retries: Optional[int] = None,
        group_name: Optional[str] = None,
        **kwargs,
    ) -> List[RetryingFuture]:
        inputs = list(map_iterdata)
        futures_list = self.executor.map(
            map_function,
            inputs,
            **kwargs,
        )
        return [
            RetryingFuture(
                f,
                map_function=map_function,
                input=i,
                map_kwargs=kwargs,
                retries=retries,
                group_name=group_name,
            )
            for i, f in zip(inputs, futures_list)
        ]

    def wait(
        self,
        fs: List[RetryingFuture],
        throw_except: Optional[bool] = True,
        return_when: Optional[Any] = ALL_COMPLETED,
        download_results: Optional[bool] = False,
        timeout: Optional[int] = None,
        threadpool_size: Optional[int] = THREADPOOL_SIZE,
        wait_dur_sec: Optional[int] = WAIT_DUR_SEC,
        show_progressbar: Optional[bool] = True,
    ) -> Tuple[List[RetryingFuture], List[RetryingFuture]]:
        lookup = {f.response_future: f for f in fs}

        while True:
            response_futures = [f.response_future for f in fs]

            done, pending = self.executor.wait(
                response_futures,
                throw_except=throw_except,
                return_when=return_when,
                download_results=download_results,
                timeout=timeout,
                threadpool_size=threadpool_size,
                wait_dur_sec=wait_dur_sec,
                show_progressbar=show_progressbar,
            )

            retrying_done = []
            retrying_pending = [lookup[response_future] for response_future in pending]
            for response_future in done:
                retrying_future = lookup[response_future]
                if response_future.error:
                    retrying_future._inc_failure_count()
                    if retrying_future._should_retry():
                        retrying_future._retry(self.executor)
                        # put back into pending since we are retrying this input
                        retrying_pending.append(retrying_future)
                        lookup[retrying_future.response_future] = retrying_future
                    else:
                        retrying_done.append(retrying_future)
                else:
                    retrying_done.append(retrying_future)

            if return_when == ALWAYS:
                break
            elif return_when == ANY_COMPLETED and len(retrying_done) > 0:
                break
            elif return_when == ALL_COMPLETED and len(retrying_pending) == 0:
                break

        return retrying_done, retrying_pending



from functools import partial

import networkx as nx
import numpy as np
import pytest
from numpy.testing import assert_array_equal

import cubed
import cubed.array_api as xp
import cubed.random
from cubed.backend_array_api import namespace as nxp
from cubed.core.ops import elemwise, merge_chunks_new, partial_reduce
from cubed.core.optimization import (
    fuse_all_optimize_dag,
    fuse_only_optimize_dag,
    fuse_predecessors,
    gensym,
    multiple_inputs_optimize_dag,
    simple_optimize_dag,
)
from cubed.core.plan import arrays_to_plan
from cubed.tests.utils import TaskCounter


@pytest.fixture()
def spec(tmp_path):
    return cubed.Spec(tmp_path, allowed_mem=100000)


def test_fusion(spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.negative(a)
    c = xp.astype(b, np.float32)
    d = xp.negative(c)

    num_arrays = 4  # a, b, c, d
    num_created_arrays = 3  # b, c, d (a is not created on disk)
    assert d.plan.num_arrays(optimize_graph=False) == num_arrays
    assert d.plan.num_tasks(optimize_graph=False) == num_created_arrays + 12
    assert (
        d.plan.total_nbytes_written(optimize_graph=False)
        == b.nbytes + c.nbytes + d.nbytes
    )
    num_arrays = 2  # a, d
    num_created_arrays = 1  # d (a is not created on disk)
    assert d.plan.num_arrays(optimize_graph=True) == num_arrays
    assert d.plan.num_tasks(optimize_graph=True) == num_created_arrays + 4
    assert d.plan.total_nbytes_written(optimize_graph=True) == d.nbytes

    task_counter = TaskCounter()
    result = d.compute(callbacks=[task_counter])
    assert task_counter.value == num_created_arrays + 4

    assert_array_equal(
        result,
        np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]).astype(np.float32),
    )


def test_fusion_transpose(spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.negative(a)
    c = xp.astype(b, np.float32)
    d = c.T

    num_created_arrays = 3  # b, c, d
    assert d.plan.num_tasks(optimize_graph=False) == num_created_arrays + 12
    num_created_arrays = 1  # d
    assert d.plan.num_tasks(optimize_graph=True) == num_created_arrays + 4

    task_counter = TaskCounter()
    result = d.compute(callbacks=[task_counter])
    assert task_counter.value == num_created_arrays + 4

    assert_array_equal(
        result,
        np.array([[-1, -4, -7], [-2, -5, -8], [-3, -6, -9]]).astype(np.float32),
    )


def test_no_fusion(spec):
    # b can't be fused with c because d also depends on b
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.positive(a)
    c = xp.positive(b)
    d = xp.equal(b, c)

    opt_fn = simple_optimize_dag

    num_created_arrays = 3  # b, c, d
    assert d.plan.num_tasks(optimize_graph=False) == num_created_arrays + 3
    assert d.plan.num_tasks(optimize_function=opt_fn) == num_created_arrays + 3

    task_counter = TaskCounter()
    result = d.compute(optimize_function=opt_fn, callbacks=[task_counter])
    assert task_counter.value == num_created_arrays + 3

    assert_array_equal(result, np.ones((2, 2)))


def test_no_fusion_multiple_edges(spec):
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.positive(a)
    c = xp.asarray(b)
    # b and c are the same array, so d has a single dependency
    # with multiple edges
    # this should not be fused under the current logic
    d = xp.equal(b, c)

    opt_fn = simple_optimize_dag

    num_created_arrays = 2  # c, d
    assert d.plan.num_tasks(optimize_graph=False) == num_created_arrays + 2
    assert d.plan.num_tasks(optimize_function=opt_fn) == num_created_arrays + 2

    task_counter = TaskCounter()
    result = d.compute(optimize_function=opt_fn, callbacks=[task_counter])
    assert task_counter.value == num_created_arrays + 2

    assert_array_equal(result, np.full((2, 2), True))


def test_custom_optimize_function(spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.negative(a)
    c = xp.astype(b, np.float32)
    d = xp.negative(c)

    num_tasks_with_no_optimization = d.plan.num_tasks(optimize_graph=False)

    assert d.plan.num_tasks(optimize_graph=True) < num_tasks_with_no_optimization

    def custom_optimize_function(dag):
        # leave DAG unchanged
        return dag

    assert (
        d.plan.num_tasks(
            optimize_graph=True, optimize_function=custom_optimize_function
        )
        == num_tasks_with_no_optimization
    )


def get_num_input_blocks(dag, arr_name):
    op_name = next(dag.predecessors(arr_name))
    return dag.nodes(data=True)[op_name]["pipeline"].config.num_input_blocks


def fuse_one_level(arr, *, always_fuse=None):
    # use fuse_predecessors to test one level of fusion
    return partial(
        fuse_predecessors,
        name=next(arr.plan.dag.predecessors(arr.name)),
        always_fuse=always_fuse,
    )


def fuse_multiple_levels(*, max_total_source_arrays=4, max_total_num_input_blocks=None):
    # use multiple_inputs_optimize_dag to test multiple levels of fusion
    return partial(
        multiple_inputs_optimize_dag,
        max_total_source_arrays=max_total_source_arrays,
        max_total_num_input_blocks=max_total_num_input_blocks,
    )


# utility functions for testing structural equivalence of dags


def create_dag():
    return nx.MultiDiGraph()


def add_op(dag, func, inputs, outputs, fusable=True):
    name = gensym(func.__name__)
    dag.add_node(name, func=func, fusable=fusable)
    for n in inputs:
        dag.add_edge(n, name)
    for n in outputs:
        dag.add_node(n)
        dag.add_edge(name, n)

    return name


def placeholder_func(*x):
    return 1


def add_placeholder_op(dag, inputs, outputs):
    add_op(dag, placeholder_func, [a.name for a in inputs], [b.name for b in outputs])


def structurally_equivalent(dag1, dag2):
    # compare structure, and node labels for values but not operators since they are placeholders

    # draw_dag(dag1, "dag1")  # uncomment for debugging
    # draw_dag(dag2, "dag2")  # uncomment for debugging

    labelled_dag1 = nx.convert_node_labels_to_integers(dag1, label_attribute="label")
    labelled_dag2 = nx.convert_node_labels_to_integers(dag2, label_attribute="label")

    def nm(node_attrs1, node_attrs2):
        label1 = node_attrs1["label"]
        label2 = node_attrs2["label"]
        # - in a label indicates that the node is an operator; don't compare these names
        if "-" in label1:
            return "-" in label2
        return label1 == label2

    return nx.is_isomorphic(labelled_dag1, labelled_dag2, node_match=nm)


def draw_dag(dag, name="dag"):
    dag = dag.copy()
    for _, d in dag.nodes(data=True):
        if "name" in d:  # pydot already has name
            del d["name"]
    gv = nx.drawing.nx_pydot.to_pydot(dag)
    format = "svg"
    full_filename = f"{name}.{format}"
    gv.write(full_filename, format=format)


# simple unary function
#
#  a        ->       a
#  |                 |
#  b                 c
#  |
#  c
#
def test_fuse_unary_op(spec):
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.negative(a)
    c = xp.negative(b)

    opt_fn = fuse_one_level(c)

    c.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (a,), (c,))
    optimized_dag = c.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(c.plan.dag, c.name) == (1,)
    assert get_num_input_blocks(optimized_dag, c.name) == (1,)

    num_created_arrays = 2  # b, c
    assert c.plan.num_tasks(optimize_graph=False) == num_created_arrays + 2
    num_created_arrays = 1  # c
    assert c.plan.num_tasks(optimize_function=opt_fn) == num_created_arrays + 1

    task_counter = TaskCounter()
    result = c.compute(callbacks=[task_counter], optimize_function=opt_fn)
    assert task_counter.value == num_created_arrays + 1

    assert_array_equal(result, np.ones((2, 2)))


# simple binary function
#
#  a   b    ->   a   b
#  |   |          \ /
#  c   d           e
#   \ /
#    e
#
def test_fuse_binary_op(spec):
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    c = xp.negative(a)
    d = xp.negative(b)
    e = xp.add(c, d)

    opt_fn = fuse_one_level(e)

    e.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (), (b,))
    add_placeholder_op(expected_fused_dag, (a, b), (e,))
    optimized_dag = e.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(e.plan.dag, e.name) == (1, 1)
    assert get_num_input_blocks(optimized_dag, e.name) == (1, 1)

    num_created_arrays = 3  # c, d, e
    assert e.plan.num_tasks(optimize_graph=False) == num_created_arrays + 3
    num_created_arrays = 1  # e
    assert e.plan.num_tasks(optimize_function=opt_fn) == num_created_arrays + 1

    task_counter = TaskCounter()
    result = e.compute(callbacks=[task_counter], optimize_function=opt_fn)
    assert task_counter.value == num_created_arrays + 1

    assert_array_equal(result, -2 * np.ones((2, 2)))


# unary and binary functions
#
#  a b   c    ->   a  b  c
#  |  \ /           \ | /
#  d   e              f
#   \ /
#    f
#
def test_fuse_unary_and_binary_op(spec):
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    c = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    d = xp.negative(a)
    e = xp.add(b, c)
    f = xp.add(d, e)

    opt_fn = fuse_one_level(f)

    f.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (), (b,))
    add_placeholder_op(expected_fused_dag, (), (c,))
    add_placeholder_op(expected_fused_dag, (a, b, c), (f,))
    optimized_dag = f.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(f.plan.dag, f.name) == (1, 1)
    assert get_num_input_blocks(optimized_dag, f.name) == (1, 1, 1)

    result = f.compute(optimize_function=opt_fn)
    assert_array_equal(result, np.ones((2, 2)))


# mixed levels
#
#    b   c    ->   a  b  c
#     \ /           \ | /
#  a   d              e
#   \ /
#    e
#
def test_fuse_mixed_levels(spec):
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    c = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    d = xp.add(b, c)
    e = xp.add(a, d)

    opt_fn = fuse_one_level(e)

    e.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (), (b,))
    add_placeholder_op(expected_fused_dag, (), (c,))
    add_placeholder_op(expected_fused_dag, (a, b, c), (e,))
    optimized_dag = e.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(e.plan.dag, e.name) == (1, 1)
    assert get_num_input_blocks(optimized_dag, e.name) == (1, 1, 1)

    result = e.compute(optimize_function=opt_fn)
    assert_array_equal(result, 3 * np.ones((2, 2)))


# diamond
#
#   a    ->    a
#  / \         ‖
# b   c        d
#  \ /
#   d
#
def test_fuse_diamond(spec):
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.positive(a)
    c = xp.positive(a)
    d = xp.add(b, c)

    opt_fn = fuse_one_level(d)

    d.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (a, a), (d,))
    optimized_dag = d.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(d.plan.dag, d.name) == (1, 1)
    assert get_num_input_blocks(optimized_dag, d.name) == (1, 1)

    result = d.compute(optimize_function=opt_fn)
    assert_array_equal(result, 2 * np.ones((2, 2)))


# mixed levels and diamond
# from https://github.com/cubed-dev/cubed/issues/126
#
#   a    ->    a
#   |         /|
#   b        b |
#  /|         \|
# c |          d
#  \|
#   d
#
def test_fuse_mixed_levels_and_diamond(spec):
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.positive(a)
    c = xp.positive(b)
    d = xp.add(b, c)

    opt_fn = fuse_one_level(d)

    d.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (a,), (b,))
    add_placeholder_op(expected_fused_dag, (a, b), (d,))
    optimized_dag = d.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(d.plan.dag, d.name) == (1, 1)
    assert get_num_input_blocks(optimized_dag, d.name) == (1, 1)

    result = d.compute(optimize_function=opt_fn)
    assert_array_equal(result, 2 * np.ones((2, 2)))


# derived from a bug found by array_api_tests/test_manipulation_functions.py::test_expand_dims
#  a   b    ->   a   b
#   \ /          |\ /|
#    c           | d |
#   /|           | | |
#  d |           | e |
#  | |            \|/
#  e |             f
#   \|
#    f
def test_fuse_mixed_levels_and_diamond_complex(spec):
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    d = xp.positive(c)
    e = d[1:, :]  # operation can't be fused
    f = xp.add(e, c)  # this order exposed a bug in argument ordering

    opt_fn = multiple_inputs_optimize_dag

    f.visualize(optimize_function=opt_fn)
    result = f.compute(optimize_function=opt_fn)
    assert_array_equal(result, 4 * np.ones((2, 2)))


# repeated argument
# from https://github.com/cubed-dev/cubed/issues/65
#
#  a   ->   a
#  |        ‖
#  b        c
#  ‖
#  c
#
def test_fuse_repeated_argument(spec):
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.negative(a)
    c = xp.add(b, b)

    opt_fn = fuse_one_level(c)

    c.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (a, a), (c,))
    optimized_dag = c.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(c.plan.dag, c.name) == (1, 1)
    assert get_num_input_blocks(optimized_dag, c.name) == (1, 1)

    result = c.compute(optimize_function=opt_fn)
    assert_array_equal(result, -2 * np.ones((2, 2)))


# other dependents
#
#   a    ->    a
#   |         / \
#   b        c   b
#  / \           |
# c   d          d
#
def test_fuse_other_dependents(spec):
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.negative(a)
    c = xp.negative(b)
    d = xp.negative(b)

    # only fuse c; leave d unfused
    opt_fn = fuse_one_level(c)

    # note multi-arg forms of visualize and compute below
    cubed.visualize(c, d, optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (a,), (b,))
    add_placeholder_op(expected_fused_dag, (a,), (c,))
    add_placeholder_op(expected_fused_dag, (b,), (d,))
    plan = arrays_to_plan(c, d)
    optimized_dag = plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(c.plan.dag, c.name) == (1,)
    assert get_num_input_blocks(optimized_dag, c.name) == (1,)

    c_result, d_result = cubed.compute(c, d, optimize_function=opt_fn)
    assert_array_equal(c_result, np.ones((2, 2)))
    assert_array_equal(d_result, np.ones((2, 2)))


# unary large fan-in
#
#  a b c d         e f g h   ->   a b c d         e f g h
#   \ \ \ \       / / / /          \ \ \ \       / / / /
#    \ \ \ \     / / / /            \ \ \ \     / / / /
#     \ \ \ \   / / / /              \ \ \ \   / / / /
#      \ \ \ \ / / / /                \ \ \ \ / / / /
#       ----- i -----                  ----- j -----
#             |
#             j
#
def test_fuse_unary_large_fan_in(spec):
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    c = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    d = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    e = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    f = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    g = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    h = xp.ones((2, 2), chunks=(2, 2), spec=spec)

    # use elemwise and stack since add can only take 2 args
    def stack_add(*a):
        return nxp.sum(nxp.stack(a), axis=0)

    i = elemwise(stack_add, a, b, c, d, e, f, g, h, dtype=a.dtype)
    j = xp.negative(i)

    # max_total_source_arrays is left at its default (4) which does not limit fusion since j is unary
    opt_fn = fuse_one_level(j)

    j.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (), (b,))
    add_placeholder_op(expected_fused_dag, (), (c,))
    add_placeholder_op(expected_fused_dag, (), (d,))
    add_placeholder_op(expected_fused_dag, (), (e,))
    add_placeholder_op(expected_fused_dag, (), (f,))
    add_placeholder_op(expected_fused_dag, (), (g,))
    add_placeholder_op(expected_fused_dag, (), (h,))
    add_placeholder_op(
        expected_fused_dag,
        (
            a,
            b,
            c,
            d,
            e,
            f,
            g,
            h,
        ),
        (j,),
    )
    optimized_dag = j.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(j.plan.dag, j.name) == (1,)
    assert get_num_input_blocks(optimized_dag, j.name) == (1,) * 8

    result = j.compute(optimize_function=opt_fn)
    assert_array_equal(result, -8 * np.ones((2, 2)))


# large fan-in default
#
#  a   b c   d e   f g   h   ->   a  b   c  d e  f   g  h
#   \ /   \ /   \ /   \ /          \  \ /  /   \  \ /  /
#    i     j     k     m            -- n --     -- o --
#     \   /       \   /                 \         /
#       n           o                    \       /
#        \         /                      -- p --
#         \       /
#          \     /
#             p
#
def test_fuse_large_fan_in_default(spec):
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    c = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    d = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    e = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    f = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    g = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    h = xp.ones((2, 2), chunks=(2, 2), spec=spec)

    i = xp.add(a, b)
    j = xp.add(c, d)
    k = xp.add(e, f)
    m = xp.add(g, h)

    n = xp.add(i, j)
    o = xp.add(k, m)

    p = xp.add(n, o)

    # max_total_source_arrays is left at its default (4) so only one level is fused
    opt_fn = fuse_multiple_levels()

    p.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (), (b,))
    add_placeholder_op(expected_fused_dag, (), (c,))
    add_placeholder_op(expected_fused_dag, (), (d,))
    add_placeholder_op(expected_fused_dag, (), (e,))
    add_placeholder_op(expected_fused_dag, (), (f,))
    add_placeholder_op(expected_fused_dag, (), (g,))
    add_placeholder_op(expected_fused_dag, (), (h,))
    add_placeholder_op(expected_fused_dag, (a, b, c, d), (n,))
    add_placeholder_op(expected_fused_dag, (e, f, g, h), (o,))
    add_placeholder_op(expected_fused_dag, (n, o), (p,))
    optimized_dag = p.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(p.plan.dag, p.name) == (1, 1)
    assert get_num_input_blocks(optimized_dag, p.name) == (1, 1)

    result = p.compute(optimize_function=opt_fn)
    assert_array_equal(result, 8 * np.ones((2, 2)))


# large fan-in override
#
#  a   b c   d e   f g   h   ->   a b c d         e f g h
#   \ /   \ /   \ /   \ /          \ \ \ \       / / / /
#    i     j     k     m            \ \ \ \     / / / /
#     \   /       \   /              \ \ \ \   / / / /
#       n           o                 \ \ \ \ / / / /
#        \         /                   ----- p -----
#         \       /
#          \     /
#             p
#
def test_fuse_large_fan_in_override(spec):
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    c = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    d = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    e = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    f = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    g = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    h = xp.ones((2, 2), chunks=(2, 2), spec=spec)

    i = xp.add(a, b)
    j = xp.add(c, d)
    k = xp.add(e, f)
    m = xp.add(g, h)

    n = xp.add(i, j)
    o = xp.add(k, m)

    p = xp.add(n, o)

    # max_total_source_arrays is overridden so multiple levels are fused
    opt_fn = fuse_multiple_levels(max_total_source_arrays=8)

    p.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (), (b,))
    add_placeholder_op(expected_fused_dag, (), (c,))
    add_placeholder_op(expected_fused_dag, (), (d,))
    add_placeholder_op(expected_fused_dag, (), (e,))
    add_placeholder_op(expected_fused_dag, (), (f,))
    add_placeholder_op(expected_fused_dag, (), (g,))
    add_placeholder_op(expected_fused_dag, (), (h,))
    add_placeholder_op(
        expected_fused_dag,
        (
            a,
            b,
            c,
            d,
            e,
            f,
            g,
            h,
        ),
        (p,),
    )
    optimized_dag = p.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(p.plan.dag, p.name) == (1, 1)
    assert get_num_input_blocks(optimized_dag, p.name) == (1,) * 8

    result = p.compute(optimize_function=opt_fn)
    assert_array_equal(result, 8 * np.ones((2, 2)))

    # now force everything to be fused with fuse_all_optimize_dag
    # note that max_total_source_arrays is *not* set
    opt_fn = fuse_all_optimize_dag
    optimized_dag = p.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)

    result = p.compute(optimize_function=opt_fn)
    assert_array_equal(result, 8 * np.ones((2, 2)))


# merge chunks with same number of tasks (unary)
#
#  a        ->       a
#  | 3               | 3
#  b                 c
#  | 1
#  c
#
def test_fuse_with_merge_chunks_unary(spec):
    a = xp.ones((3, 2), chunks=(1, 2), spec=spec)
    b = merge_chunks_new(a, chunks=(3, 2))
    c = xp.negative(b)

    opt_fn = fuse_one_level(c)

    c.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (a,), (c,))
    optimized_dag = c.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(b.plan.dag, b.name) == (3,)
    assert get_num_input_blocks(c.plan.dag, c.name) == (1,)
    assert get_num_input_blocks(optimized_dag, c.name) == (3,)

    result = c.compute(optimize_function=opt_fn)
    assert_array_equal(result, -np.ones((3, 2)))


# merge chunks with same number of tasks (binary)
#
#   a   b    ->   a   b
# 3 |   | 1      3 \ / 1
#   c   d           e
#  1 \ / 1
#     e
#
def test_fuse_with_merge_chunks_binary(spec):
    a = xp.ones((3, 2), chunks=(1, 2), spec=spec)
    b = xp.ones((3, 2), chunks=(3, 2), spec=spec)
    c = merge_chunks_new(a, chunks=(3, 2))
    d = xp.negative(b)
    e = xp.add(c, d)

    opt_fn = fuse_one_level(e)

    e.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (), (b,))
    add_placeholder_op(expected_fused_dag, (a, b), (e,))
    optimized_dag = e.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(e.plan.dag, e.name) == (1, 1)
    assert get_num_input_blocks(optimized_dag, e.name) == (3, 1)

    result = e.compute(optimize_function=opt_fn)
    assert_array_equal(result, np.zeros((3, 2)))


# merge chunks with different number of tasks (b has more tasks than c)
#
#  a        ->       a
#  | 1               | 3
#  b                 c
#  | 3
#  c
#
def test_fuse_merge_chunks_unary(spec):
    a = xp.ones((3, 2), chunks=(1, 2), spec=spec)
    b = xp.negative(a)
    c = merge_chunks_new(b, chunks=(3, 2))

    # specify max_total_num_input_blocks to force c to fuse
    opt_fn = fuse_multiple_levels(max_total_num_input_blocks=3)

    c.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (a,), (c,))
    optimized_dag = c.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(b.plan.dag, b.name) == (1,)
    assert get_num_input_blocks(c.plan.dag, c.name) == (3,)
    assert get_num_input_blocks(optimized_dag, c.name) == (3,)

    result = c.compute(optimize_function=opt_fn)
    assert_array_equal(result, -np.ones((3, 2)))


# merge chunks with different number of tasks (c has more tasks than d)
#
#  a   b    ->   a   b
# 1 \ / 1       3 \ / 3
#    c             d
#    | 3
#    d
#
def test_fuse_merge_chunks_binary(spec):
    a = xp.ones((3, 2), chunks=(1, 2), spec=spec)
    b = xp.ones((3, 2), chunks=(1, 2), spec=spec)
    c = xp.add(a, b)
    d = merge_chunks_new(c, chunks=(3, 2))

    # specify max_total_num_input_blocks to force d to fuse
    opt_fn = fuse_multiple_levels(max_total_num_input_blocks=6)

    d.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (), (b,))
    add_placeholder_op(expected_fused_dag, (a, b), (d,))
    optimized_dag = d.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(c.plan.dag, c.name) == (1, 1)
    assert get_num_input_blocks(d.plan.dag, d.name) == (3,)
    assert get_num_input_blocks(optimized_dag, d.name) == (3, 3)

    result = d.compute(optimize_function=opt_fn)
    assert_array_equal(result, 2 * np.ones((3, 2)))


# like test_fuse_merge_chunks_unary, except uses partial_reduce
def test_fuse_partial_reduce_unary(spec):
    a = xp.ones((3, 2), chunks=(1, 2), spec=spec)
    b = xp.negative(a)
    c = partial_reduce(b, np.sum, split_every={0: 3})

    # specify max_total_num_input_blocks to force c to fuse
    opt_fn = fuse_multiple_levels(max_total_num_input_blocks=3)

    c.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (a,), (c,))
    optimized_dag = c.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(b.plan.dag, b.name) == (1,)
    assert get_num_input_blocks(c.plan.dag, c.name) == (3,)
    assert get_num_input_blocks(optimized_dag, c.name) == (3,)

    result = c.compute(optimize_function=opt_fn)
    assert_array_equal(result, -3 * np.ones((1, 2)))


# like test_fuse_merge_chunks_binary, except uses partial_reduce
def test_fuse_partial_reduce_binary(spec):
    a = xp.ones((3, 2), chunks=(1, 2), spec=spec)
    b = xp.ones((3, 2), chunks=(1, 2), spec=spec)
    c = xp.add(a, b)
    d = partial_reduce(c, np.sum, split_every={0: 3})

    # specify max_total_num_input_blocks to force d to fuse
    opt_fn = fuse_multiple_levels(max_total_num_input_blocks=6)

    d.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (), (b,))
    add_placeholder_op(expected_fused_dag, (a, b), (d,))
    optimized_dag = d.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(c.plan.dag, c.name) == (1, 1)
    assert get_num_input_blocks(d.plan.dag, d.name) == (3,)
    assert get_num_input_blocks(optimized_dag, d.name) == (3, 3)

    result = d.compute(optimize_function=opt_fn)
    assert_array_equal(result, 6 * np.ones((1, 2)))


def test_fuse_only_optimize_dag(spec):
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.negative(a)
    c = xp.negative(b)
    d = xp.negative(c)

    # only fuse d (with c)
    # b should remain un-fused, even though it is fusable
    op_name = next(d.plan.dag.predecessors(d.name))
    opt_fn = partial(fuse_only_optimize_dag, only_fuse=[op_name])

    c.visualize(optimize_function=opt_fn)

    # check structure of optimized dag
    expected_fused_dag = create_dag()
    add_placeholder_op(expected_fused_dag, (), (a,))
    add_placeholder_op(expected_fused_dag, (a,), (b,))
    add_placeholder_op(expected_fused_dag, (b,), (d,))
    optimized_dag = d.plan.optimize(optimize_function=opt_fn).dag
    assert structurally_equivalent(optimized_dag, expected_fused_dag)
    assert get_num_input_blocks(d.plan.dag, d.name) == (1,)
    assert get_num_input_blocks(optimized_dag, d.name) == (1,)

    result = d.compute(optimize_function=opt_fn)
    assert_array_equal(result, -np.ones((2, 2)))


def test_optimize_stack(spec):
    # This test fails if stack's general_blockwise call doesn't have fusable=False
    a = cubed.random.random((10, 10), chunks=(5, 5), spec=spec)
    b = cubed.random.random((10, 10), chunks=(5, 5), spec=spec)
    c = xp.stack((a, b), axis=0)
    # try to fuse all ops into one
    c.compute(optimize_function=fuse_multiple_levels(max_total_num_input_blocks=10))



import inspect
import tempfile
import uuid
from datetime import datetime
from functools import lru_cache
from typing import Callable, Optional

import networkx as nx
import zarr

from cubed.core.optimization import multiple_inputs_optimize_dag
from cubed.primitive.blockwise import BlockwiseSpec
from cubed.primitive.types import PrimitiveOperation
from cubed.runtime.pipeline import visit_nodes
from cubed.runtime.types import ComputeEndEvent, ComputeStartEvent, CubedPipeline
from cubed.storage.zarr import LazyZarrArray
from cubed.utils import chunk_memory, extract_stack_summaries, join_path, memory_repr

# A unique ID with sensible ordering, used for making directory names
CONTEXT_ID = f"cubed-{datetime.now().strftime('%Y%m%dT%H%M%S')}-{uuid.uuid4()}"

sym_counter = 0


def gensym(name="op"):
    global sym_counter
    sym_counter += 1
    return f"{name}-{sym_counter:03}"


class Plan:
    """Deferred computation plan for a graph of arrays.

    A thin wrapper around a NetworkX `MultiDiGraph`. Nodes are Zarr paths, and may
    have a `pipeline` attribute holding the pipeline to execute to generate
    the output at that path. If there is no `pipeline` attribute no computation
    is needed since the Zarr file already exists. Directed edges point towards
    the dependent Zarr path.

    Multiple edges are possible since a node (Zarr file) may be computed by
    a function with repeated inputs. For example, consider `equals` where the
    two arguments are the same array. We need to keep track of these cases, so
    we use a NetworkX `MultiDiGraph` rather than just as `DiGraph`.

    Compared to a more traditional DAG representing a computation, in Cubed
    nodes are not values that are passed to functions, they are instead
    "parallel computations" which are run for their side effects. Data does
    not flow through the graph - it is written to external storage (Zarr files)
    as the output of one pipeline, then read back as the input to later pipelines.
    """

    def __init__(self, dag):
        self.dag = dag

    # args from pipeline onwards are omitted for creation functions when no computation is needed
    @classmethod
    def _new(
        cls,
        name,
        op_name,
        target,
        primitive_op=None,
        hidden=False,
        *source_arrays,
    ):
        # create an empty DAG or combine from sources
        if len(source_arrays) == 0:
            dag = nx.MultiDiGraph()
        else:
            dag = arrays_to_dag(*source_arrays)

        # add new node and edges
        frame = inspect.currentframe().f_back  # go back one in the stack
        stack_summaries = extract_stack_summaries(frame, limit=10)

        first_cubed_i = min(i for i, s in enumerate(stack_summaries) if s.is_cubed())
        first_cubed_summary = stack_summaries[first_cubed_i]

        op_name_unique = gensym()

        if primitive_op is None:
            # op
            dag.add_node(
                op_name_unique,
                name=op_name_unique,
                op_name=op_name,
                type="op",
                stack_summaries=stack_summaries,
                op_display_name=f"{op_name_unique}\n{first_cubed_summary.name}",
                hidden=hidden,
            )
            # array (when multiple outputs are supported there could be more than one)
            dag.add_node(
                name,
                name=name,
                type="array",
                target=target,
                hidden=hidden,
            )
            dag.add_edge(op_name_unique, name)
        else:
            # op
            dag.add_node(
                op_name_unique,
                name=op_name_unique,
                op_name=op_name,
                type="op",
                stack_summaries=stack_summaries,
                op_display_name=f"{op_name_unique}\n{first_cubed_summary.name}",
                hidden=hidden,
                primitive_op=primitive_op,
                pipeline=primitive_op.pipeline,
            )
            # array (when multiple outputs are supported there could be more than one)
            dag.add_node(
                name,
                name=name,
                type="array",
                target=target,
                hidden=hidden,
            )
            dag.add_edge(op_name_unique, name)
        for x in source_arrays:
            if hasattr(x, "name"):
                dag.add_edge(x.name, op_name_unique)

        return Plan(dag)

    @classmethod
    def arrays_to_plan(cls, *arrays):
        return Plan(arrays_to_dag(*arrays))

    def optimize(
        self,
        optimize_function: Optional[Callable[..., nx.MultiDiGraph]] = None,
    ):
        if optimize_function is None:
            optimize_function = multiple_inputs_optimize_dag
        dag = optimize_function(self.dag)
        return Plan(dag)

    def _create_lazy_zarr_arrays(self, dag):
        # find all lazy zarr arrays in dag
        all_pipeline_nodes = []
        lazy_zarr_arrays = []
        allowed_mem = 0
        reserved_mem = 0
        for n, d in dag.nodes(data=True):
            if "primitive_op" in d:
                all_pipeline_nodes.append(n)
                allowed_mem = max(allowed_mem, d["primitive_op"].allowed_mem)
                reserved_mem = max(reserved_mem, d["primitive_op"].reserved_mem)

            if "target" in d and isinstance(d["target"], LazyZarrArray):
                lazy_zarr_arrays.append(d["target"])

        if len(lazy_zarr_arrays) > 0:
            # add new node and edges
            name = "create-arrays"
            op_name = name
            primitive_op = create_zarr_arrays(
                lazy_zarr_arrays, allowed_mem, reserved_mem
            )
            dag.add_node(
                name,
                name=name,
                op_name=op_name,
                type="op",
                op_display_name=name,
                primitive_op=primitive_op,
                pipeline=primitive_op.pipeline,
            )
            dag.add_node(
                "arrays",
                name="arrays",
                target=None,
            )
            dag.add_edge(name, "arrays")
            # make create arrays node a predecessor of all pipeline nodes so it runs first
            for n in all_pipeline_nodes:
                dag.add_edge("arrays", n)

        return dag

    @lru_cache
    def _finalize_dag(
        self, optimize_graph: bool = True, optimize_function=None
    ) -> nx.MultiDiGraph:
        dag = self.optimize(optimize_function).dag if optimize_graph else self.dag
        # create a copy since _create_lazy_zarr_arrays mutates the dag
        dag = dag.copy()
        dag = self._create_lazy_zarr_arrays(dag)
        return nx.freeze(dag)

    def execute(
        self,
        executor=None,
        callbacks=None,
        optimize_graph=True,
        optimize_function=None,
        resume=None,
        spec=None,
        **kwargs,
    ):
        dag = self._finalize_dag(optimize_graph, optimize_function)

        compute_id = f"compute-{datetime.now().strftime('%Y%m%dT%H%M%S.%f')}"

        if callbacks is not None:
            event = ComputeStartEvent(compute_id, dag, resume)
            [callback.on_compute_start(event) for callback in callbacks]
        executor.execute_dag(
            dag,
            compute_id=compute_id,
            callbacks=callbacks,
            resume=resume,
            spec=spec,
            **kwargs,
        )
        if callbacks is not None:
            event = ComputeEndEvent(compute_id, dag)
            [callback.on_compute_end(event) for callback in callbacks]

    def num_tasks(self, optimize_graph=True, optimize_function=None, resume=None):
        """Return the number of tasks needed to execute this plan."""
        dag = self._finalize_dag(optimize_graph, optimize_function)
        tasks = 0
        for _, node in visit_nodes(dag, resume=resume):
            tasks += node["primitive_op"].num_tasks
        return tasks

    def num_arrays(self, optimize_graph: bool = True, optimize_function=None) -> int:
        """Return the number of arrays in this plan."""
        dag = self._finalize_dag(optimize_graph, optimize_function)
        return sum(d.get("type") == "array" for _, d in dag.nodes(data=True))

    def max_projected_mem(
        self, optimize_graph=True, optimize_function=None, resume=None
    ):
        """Return the maximum projected memory across all tasks to execute this plan."""
        dag = self._finalize_dag(optimize_graph, optimize_function)
        projected_mem_values = [
            node["primitive_op"].projected_mem
            for _, node in visit_nodes(dag, resume=resume)
        ]
        return max(projected_mem_values) if len(projected_mem_values) > 0 else 0

    def total_nbytes_written(
        self, optimize_graph: bool = True, optimize_function=None
    ) -> int:
        """Return the total number of bytes written for all materialized arrays in this plan."""
        dag = self._finalize_dag(optimize_graph, optimize_function)
        nbytes = 0
        for _, d in dag.nodes(data=True):
            if d.get("type") == "array":
                target = d["target"]
                if isinstance(target, LazyZarrArray):
                    nbytes += target.nbytes
        return nbytes

    def visualize(
        self,
        filename="cubed",
        format=None,
        rankdir="TB",
        optimize_graph=True,
        optimize_function=None,
        show_hidden=False,
    ):
        dag = self._finalize_dag(optimize_graph, optimize_function)
        dag = dag.copy()  # make a copy since we mutate the DAG below

        # remove edges from create-arrays output node to avoid cluttering the diagram
        dag.remove_edges_from(list(dag.out_edges("arrays")))

        if not show_hidden:
            dag.remove_nodes_from(
                list(n for n, d in dag.nodes(data=True) if d.get("hidden", False))
            )

        dag.graph["graph"] = {
            "rankdir": rankdir,
            "label": (
                # note that \l is used to left-justify each line (see https://www.graphviz.org/docs/attrs/nojustify/)
                rf"num tasks: {self.num_tasks(optimize_graph, optimize_function)}\l"
                rf"max projected memory: {memory_repr(self.max_projected_mem(optimize_graph, optimize_function))}\l"
                rf"total nbytes written: {memory_repr(self.total_nbytes_written(optimize_graph, optimize_function))}\l"
                rf"optimized: {optimize_graph}\l"
            ),
            "labelloc": "bottom",
            "labeljust": "left",
            "fontsize": "10",
        }
        dag.graph["node"] = {"fontname": "helvetica", "shape": "box", "fontsize": "10"}

        # do an initial pass to extract array variable names from stack summaries
        array_display_names = {}
        for n, d in dag.nodes(data=True):
            if "stack_summaries" in d:
                stack_summaries = d["stack_summaries"]
                first_cubed_i = min(
                    i for i, s in enumerate(stack_summaries) if s.is_cubed()
                )
                caller_summary = stack_summaries[first_cubed_i - 1]
                array_display_names.update(caller_summary.array_names_to_variable_names)
        # add current stack info
        frame = inspect.currentframe().f_back  # go back one in the stack
        stack_summaries = extract_stack_summaries(frame, limit=10)
        first_cubed_i = min(i for i, s in enumerate(stack_summaries) if s.is_cubed())
        caller_summary = stack_summaries[first_cubed_i - 1]
        array_display_names.update(caller_summary.array_names_to_variable_names)

        # now set node attributes with visualization info
        for n, d in dag.nodes(data=True):
            label = n
            tooltip = f"name: {n}\n"
            node_type = d.get("type", None)
            if node_type == "op":
                label = d["op_display_name"]
                op_name = d["op_name"]
                if op_name == "blockwise":
                    d["style"] = '"rounded,filled"'
                    d["fillcolor"] = "#dcbeff"
                elif op_name == "rechunk":
                    d["style"] = '"rounded,filled"'
                    d["fillcolor"] = "#aaffc3"
                else:
                    # creation function
                    d["style"] = "rounded"
                tooltip += f"op: {op_name}"

                num_tasks = None
                if "primitive_op" in d:
                    primitive_op = d["primitive_op"]
                    tooltip += (
                        f"\nprojected memory: {memory_repr(primitive_op.projected_mem)}"
                    )
                    num_tasks = primitive_op.num_tasks
                    tooltip += f"\ntasks: {num_tasks}"
                    if primitive_op.write_chunks is not None:
                        tooltip += f"\nwrite chunks: {primitive_op.write_chunks}"
                    del d["primitive_op"]

                # remove pipeline attribute since it is a long string that causes graphviz to fail
                if "pipeline" in d:
                    pipeline = d["pipeline"]
                    if isinstance(pipeline.config, BlockwiseSpec):
                        tooltip += (
                            f"\nnum input blocks: {pipeline.config.num_input_blocks}"
                        )
                    del d["pipeline"]

                if "stack_summaries" in d and d["stack_summaries"] is not None:
                    # add call stack information
                    stack_summaries = d["stack_summaries"]

                    first_cubed_i = min(
                        i for i, s in enumerate(stack_summaries) if s.is_cubed()
                    )
                    caller_summary = stack_summaries[first_cubed_i - 1]

                    calls = " -> ".join(
                        [
                            s.name
                            for s in stack_summaries
                            if not s.is_on_python_lib_path()
                        ]
                    )

                    line = f"{caller_summary.lineno} in {caller_summary.name}"

                    tooltip += f"\ncalls: {calls}"
                    tooltip += f"\nline: {line}"
                    del d["stack_summaries"]

                if num_tasks is not None:
                    label += f"\ntasks: {num_tasks}"

            elif node_type == "array":
                target = d["target"]
                chunkmem = memory_repr(chunk_memory(target))
                nbytes = None

                # materialized arrays are light orange, virtual arrays are white
                if isinstance(target, (LazyZarrArray, zarr.Array)):
                    d["style"] = "filled"
                    d["fillcolor"] = "#ffd8b1"
                    nbytes = memory_repr(target.nbytes)
                if n in array_display_names:
                    var_name = array_display_names[n]
                    label = f"{n}\n{var_name}"
                    tooltip += f"variable: {var_name}\n"
                tooltip += f"shape: {target.shape}\n"
                tooltip += f"chunks: {target.chunks}\n"
                tooltip += f"dtype: {target.dtype}\n"
                tooltip += f"chunk memory: {chunkmem}\n"
                if nbytes is not None:
                    tooltip += f"nbytes: {nbytes}\n"

                del d["target"]

            # quote strings with colons in them (https://github.com/pydot/pydot/issues/258)
            d["label"] = '"' + label.strip() + '"'
            d["tooltip"] = '"' + tooltip.strip() + '"'

            if "name" in d:  # pydot already has name
                del d["name"]
        gv = nx.drawing.nx_pydot.to_pydot(dag)
        if format is None:
            format = "svg"
        full_filename = f"{filename}.{format}"
        gv.write(full_filename, format=format)

        try:  # pragma: no cover
            import IPython.display as display

            if format == "svg":
                return display.SVG(filename=full_filename)
        except ImportError:
            # Can't return a display object if no IPython.
            pass
        return None


def arrays_to_dag(*arrays):
    from .array import check_array_specs

    check_array_specs(arrays)
    dags = [x.plan.dag for x in arrays if hasattr(x, "plan")]
    return nx.compose_all(dags)


def arrays_to_plan(*arrays):
    plans = [x.plan for x in arrays if hasattr(x, "plan")]
    if len(plans) == 0:
        raise ValueError(f"No plans found for arrays: {arrays}")
    return plans[0].arrays_to_plan(*arrays)


def new_temp_path(name, suffix=".zarr", spec=None):
    work_dir = spec.work_dir if spec is not None else None
    if work_dir is None:
        work_dir = tempfile.gettempdir()
    context_dir = join_path(work_dir, CONTEXT_ID)
    return join_path(context_dir, f"{name}{suffix}")


def create_zarr_array(lazy_zarr_array, *, config=None):
    """Stage function for create."""
    lazy_zarr_array.create(mode="a")


def create_zarr_arrays(lazy_zarr_arrays, allowed_mem, reserved_mem):
    # projected memory is size of largest dtype size (for a fill value)
    projected_mem = (
        max([lza.dtype.itemsize for lza in lazy_zarr_arrays], default=0) + reserved_mem
    )
    num_tasks = len(lazy_zarr_arrays)

    pipeline = CubedPipeline(
        create_zarr_array,
        "create_zarr_array",
        lazy_zarr_arrays,
        None,
    )
    return PrimitiveOperation(
        pipeline=pipeline,
        source_array_names=[],
        target_array=None,
        projected_mem=projected_mem,
        allowed_mem=allowed_mem,
        reserved_mem=reserved_mem,
        num_tasks=num_tasks,
        fusable=False,
    )



import platform
import sys
from typing import Iterable

import networkx as nx
import numpy as np
import zarr

from cubed.runtime.create import create_executor
from cubed.runtime.types import Callback

LITHOPS_LOCAL_CONFIG = {"lithops": {"backend": "localhost", "storage": "localhost"}}

ALL_EXECUTORS = [create_executor("single-threaded")]

# don't run all tests on every executor as it's too slow, so just have a subset
MAIN_EXECUTORS = [create_executor("single-threaded")]


if platform.system() != "Windows":
    # AsyncPythonDagExecutor calls `peak_measured_mem` which is not supported on Windows
    ALL_EXECUTORS.append(create_executor("threads"))

    # AsyncPythonDagExecutor (processes) uses an API available from 3.11 onwards (max_tasks_per_child)
    if sys.version_info >= (3, 11):
        ALL_EXECUTORS.append(create_executor("processes"))
        MAIN_EXECUTORS.append(create_executor("processes"))

try:
    ALL_EXECUTORS.append(create_executor("beam"))
    MAIN_EXECUTORS.append(create_executor("beam"))
except ImportError:
    pass

try:
    ALL_EXECUTORS.append(create_executor("dask"))
    MAIN_EXECUTORS.append(create_executor("dask"))
except ImportError:
    pass

try:
    executor_options = dict(config=LITHOPS_LOCAL_CONFIG)
    ALL_EXECUTORS.append(create_executor("lithops", executor_options))
    MAIN_EXECUTORS.append(create_executor("lithops", executor_options))
except ImportError:
    pass

MODAL_EXECUTORS = []

try:
    MODAL_EXECUTORS.append(create_executor("modal"))
except ImportError:
    pass


class TaskCounter(Callback):
    def __init__(self, check_timestamps=True) -> None:
        self.check_timestamps = check_timestamps

    def on_compute_start(self, event):
        self.value = 0

    def on_task_end(self, event):
        if self.check_timestamps and event.task_create_tstamp is not None:
            assert (
                event.task_result_tstamp
                >= event.function_end_tstamp
                >= event.function_start_tstamp
                >= event.task_create_tstamp
                > 0
            )
        self.value += event.num_tasks


def create_zarr(a, /, store, *, dtype=None, chunks=None, path=None):
    # from dask.asarray
    if not isinstance(getattr(a, "shape", None), Iterable):
        # ensure blocks are arrays
        a = np.asarray(a, dtype=dtype)
    if dtype is None:
        dtype = a.dtype

    # write to zarr
    za = zarr.open(
        store, mode="w", shape=a.shape, dtype=dtype, chunks=chunks, path=path
    )
    za[:] = a
    return za


def execute_pipeline(pipeline, executor):
    """Executes a pipeline"""
    dag = nx.MultiDiGraph()
    dag.add_node("node", pipeline=pipeline)
    executor.execute_dag(dag)



from cubed.array_api.creation_functions import asarray
from cubed.backend_array_api import namespace as nxp
from cubed.core import reduction


def all(x, /, *, axis=None, keepdims=False, use_new_impl=True, split_every=None):
    if x.size == 0:
        return asarray(True, dtype=x.dtype)
    return reduction(
        x,
        nxp.all,
        axis=axis,
        dtype=bool,
        keepdims=keepdims,
        use_new_impl=use_new_impl,
        split_every=split_every,
    )


def any(x, /, *, axis=None, keepdims=False, use_new_impl=True, split_every=None):
    if x.size == 0:
        return asarray(False, dtype=x.dtype)
    return reduction(
        x,
        nxp.any,
        axis=axis,
        dtype=bool,
        keepdims=keepdims,
        use_new_impl=use_new_impl,
        split_every=split_every,
    )



import pytest

from cubed.storage.zarr import lazy_zarr_array


def test_lazy_zarr_array(tmp_path):
    zarr_path = tmp_path / "lazy.zarr"
    arr = lazy_zarr_array(zarr_path, shape=(3, 3), dtype=int, chunks=(2, 2))

    assert not zarr_path.exists()
    with pytest.raises(ValueError):
        arr.open()

    arr.create()
    assert zarr_path.exists()
    arr.open()
